<TeXmacs|1.99.2>

<style|<tuple|article|coq>>

<\body>
  \;

  <\comment>
    <doc-data|<doc-title|A verified abstract machine for<next-line>functional
    coroutines <math|-> Coq formalization<next-line>(companion technical
    report)>|<doc-author|<author-data|<author-name|Tristan
    Crolard>|<\author-affiliation>
      CNAM Paris, France
    </author-affiliation>|<\author-email>
      <\code>
        tristan.crolard@cnam.fr

        \;
      </code>
    </author-email>>>>

    \;
  </comment>

  <value|begin_comment>

  <center|<with|font-series|bold|Abstract>>

  <\quote-env>
    Functional coroutines are a restricted form of control mechanism, where
    each continuation comes with its local environment. This restriction was
    originally obtained by considering a constructive version of Parigot's
    classical natural deduction which is sound and complete for the Constant
    Domain logic. In this article, we present a refinement of de Groote's
    abstract machine which is proved to be correct for functional coroutines.
    Therefore, this abstract machine also provides a direct computational
    interpretation of the Constant Domain logic.

    \;

    This companion technical report contains the full Coq formalization (all
    the proofs are available in the Coq source file). The section numbering
    is the same as in the article for easy cross referencing.
  </quote-env>

  <value|end_comment>

  <\comment>
    <section|Introduction>

    \;

    <section|Dependency relations>

    \;

    <section|Abstract machines>

    \;
  </comment>

  <with|font-series|bold|Require> <with|font-series|bold|Import>
  <with|font-shape|italic|Utf8>.

  <with|font-series|bold|Add LoadPath> "<samp|~/Sources/Coq/Tactics>".

  <with|font-series|bold|Require> <with|font-series|bold|Import>
  <with|font-shape|italic|CpdtTactics> <with|font-shape|italic|LibTactics>
  <with|font-shape|italic|TwelfTactics>.

  \;

  <with|font-series|bold|Require Import> <with|font-shape|italic|Arith>
  <with|font-shape|italic|List> <with|font-shape|italic|String>.

  <with|font-series|bold|Require Import> <with|font-shape|italic|Coq.extraction.ExtrOcamlString>.

  \;

  <with|font-series|bold|Definition> <em|ble_nat> (<math|n> : <em|nat>)
  (<math|m> : <em|nat>) : <em|bool> := <with|font-shape|italic|beq_nat>
  (<math|n> <math|-> <math|m>) 0.

  <with|font-series|bold|Notation> "<math|n> <math|==> <math|m>" :=
  (<with|font-shape|italic|beq_nat> <math|n> <math|m>) (at level 70, no
  associativity).

  <with|font-series|bold|Notation> "<math|n> <math|\<leqq\>> <math|m>" :=
  (<with|font-shape|italic|ble_nat> <math|n> <math|m>) (at level 70, no
  associativity).

  \;

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|m<rsub|1>> <protectlr|<math|->> <math|m<rsub|2>> <protectlr|=>
  <math|m>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|minus> :
  <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \;

  \| <with|font-shape|italic|Minus<math|<rsub|1>>> : <math|\<forall\>>
  <math|m<rsub|1>>,

  <\equation*>
    <math|m<rsub|1>> <protectlr|<math|->> 0 <protectlr|=> <math|m<rsub|1>>
  </equation*>

  \;

  \| <with|font-shape|italic|Minus<math|<rsub|2>>> : <math|\<forall\>>
  <math|m<rsub|1>> <math|m<rsub|2>> <math|m>,

  <\equation*>
    <dfrac|<math|m<rsub|1>> <protectlr|<math|->> m<rsub|2> <protectlr|=>
    <math|m>|S<math|<around*|(|m<rsub|1>|)>> <protectlr|<math|->>
    S<around*|(|m<rsub|2>|)> <protectlr|=> m>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|m<rsub|1>> <protectlr|<math|->>
  <math|m<rsub|2>> <protectlr|=> <math|m>" := (<with|font-shape|italic|minus>
  <math|m<rsub|1>> <math|m<rsub|2>> <math|m>).

  \;

  <with|font-series|bold|Fixpoint> <math|f> (<math|m> :
  <with|font-shape|italic|nat>) : (<math|<math|m> <protectlr|<math|->> m
  <protectlr|=> 0>) :=

  \ \ <with|font-series|bold|match> <with|font-shape|italic|m>
  <with|font-series|bold|with>

  \ \ \| 0 <verbatim|=\<gtr\>> <with|font-shape|italic|Minus><math|<rsub|1>>
  0

  \ \ \| <math|S>(<math|n>) <verbatim|=\<gtr\>>
  <with|font-shape|italic|Minus><math|<rsub|2>> <math|n> <math|n> 0 (<math|f>
  <math|n>)

  \ \ <with|font-series|bold|end>.

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|minus_trivial> :
  <math|\<forall\>> <math|n>, <math|<math-it|not> <around*|(|0=S n|)>>.

  <\proof>
    \ \ auto.
  </proof>

  <with|font-series|bold|Lemma> <with|font-shape|italic|minus_id> :
  <math|\<forall\>> <math|m>, <math|<math|m> <protectlr|<math|->> m
  <protectlr|=> 0>.

  <\proof>
    \ \ exact f.
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|minus_id>.

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|minus_unique> :
  <math|\<forall\>> <math|m<rsub|2>> <math|m<rsub|1>> <math|m>
  <math|m<rprime|'>>, <math|<math|m<rsub|1>> <protectlr|<math|->>m<rsub|2>
  <protectlr|=> m<space|1spc>\<rightarrow\><space|1spc>m<rsub|1>
  <protectlr|<math|->> m<rsub|2> <protectlr|=>
  m<rprime|'><space|1spc>\<rightarrow\><space|1spc>m=m<rprime|'>>.

  <\proof>
    \ \ intros.

    \ \ induction H.

    \ \ - inversion H0.

    \ \ \ \ auto.

    \ \ - inversion H0.

    \ \ \ \ auto.
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|minus_unique> : <em|gen_subst_db>.

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|minus_succ> :
  <math|\<forall\>> <math|n<rsub|3>> <math|n<rsub|2>> <math|n<rsub|1>>,
  <math|<math|n<rsub|1>> <protectlr|<math|->><around*|(|S n<rsub|2>|)>
  <protectlr|=> n<rsub|3><space|1spc>\<rightarrow\><space|1spc><math|n<rsub|1>>
  <protectlr|<math|->> n<rsub|2> <protectlr|=> <around*|(|S n<rsub|3>|)>>.

  <\proof>
    \ \ intros <math|n<rsub|3>> <math|n<rsub|2>>.

    \ \ induction <math|n<rsub|2>>.

    \ \ - intros <math|n<rsub|1>> <math|\<cal-D\>>.\ 

    \ \ \ \ inversion <math|\<cal-D\>>.

    \ \ \ \ inversion H1.

    \ \ \ \ apply <math|Minus<rsub|1>>.

    \ \ - intros <math|n<rsub|1>> <math|\<cal-D\>>.\ 

    \ \ \ \ inversion <math|\<cal-D\>>.

    \ \ \ \ apply <math|Minus<rsub|2>>.

    \ \ \ \ apply <math|IHn<rsub|2>>.

    \ \ \ \ exact H1.
  </proof>

  <with|font-series|bold|Lemma> <with|font-shape|italic|minus_swap> :
  <math|\<forall\>> <math|n<rsub|3>> <math|n<rsub|2>> <math|n<rsub|1>>,
  <math|<math|n<rsub|1>> <protectlr|<math|->> n<rsub|2> <protectlr|=>
  n<rsub|3><space|1spc>\<rightarrow\><space|1spc><math|n<rsub|1>>
  <protectlr|<math|->> n<rsub|3> <protectlr|=> n<rsub|2>>.

  <\proof>
    \ \ intros n<rsub|3> n<rsub|2>.\ 

    \ \ induction n<rsub|2>.\ 

    \ \ - intros n<rsub|1> <math|\<cal-D\>>.\ 

    \ \ \ \ inversion <math|\<cal-D\>>.\ 

    \ \ \ \ apply minus_id.\ 

    \ \ - intros n<rsub|1> <math|\<cal-D\>>.\ 

    \ \ \ \ inversion <math|\<cal-D\>>.\ 

    \ \ \ \ apply minus_succ.\ 

    \ \ \ \ apply Minus<rsub|2>.\ 

    \ \ \ \ apply IHn<rsub|2>.\ 

    \ \ \ \ exact H1.
  </proof>

  \;

  <with|font-series|bold|Definition> <with|font-shape|italic|vector> :=
  <with|font-shape|italic|list> <with|font-shape|italic|nat>.

  <with|font-series|bold|Definition> <with|font-shape|italic|table> :=
  <with|font-shape|italic|list> <with|font-shape|italic|vector>.

  \;

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<cal-I\>><space|0.2spc><protectl|(><space|0.2spc><math|n><space|0.2spc><protectr|)>
  <protectlr|=> <math|m>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|fetch> {<math|A>
  : <with|font-shape|italic|Type>} : <with|font-shape|italic|list A>
  <math|\<rightarrow\>> <with|font-shape|italic|nat> <math|\<rightarrow\>> A
  <math|\<rightarrow\>> <with|font-shape|italic|Prop> :=

  \;

  \| <with|font-shape|italic|Fetch<math|<rsub|1>>> : <math|\<forall\>>
  <math|\<cal-I\>> (<math|n>:<math|A>),

  <\equation*>
    \ <around*|(|n::\<cal-I\>|)><protectl|(>0<protectr|)><protectlr|=>
    <math|n>
  </equation*>

  \;

  \| <with|font-shape|italic|Fetch<math|<rsub|2>>> : <math|\<forall\>>
  <math|\<cal-I\>> (<math|n>:<math|A>) <math|n<rsub|1>> <math|n<rsub|2>>,

  <\equation*>
    <dfrac|\<cal-I\><protectl|(>n<rsub|1><protectr|)><protectlr|=>
    <math|n<rsub|2>>|<around*|(|n::\<cal-I\>|)><protectl|(>S
    n<rsub|1><protectr|)><protectlr|=> <math|n<rsub|2>>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<cal-I\>><space|0.2spc><protectl|(><space|0.2spc><math|n><space|0.2spc><protectr|)>
  <protectlr|=> <math|m>" := (@<with|font-shape|italic|fetch> _
  <math|\<cal-I\>> <math|n> <math|m>).

  \;

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|fetch_unique> :
  <math|\<forall\>> <math|A> <math|n> <math|\<cal-I\>> (<math|y>:<math|A>)
  (<math|y<rprime|'>>:<math|A>), <math|\<cal-I\><protectl|(>n<protectr|)><protectlr|=>y>
  <math|\<rightarrow\>> <math|\<cal-I\><protectl|(>n<protectr|)><protectlr|=>
  y<rprime|'>> <math|\<rightarrow\>> <math|y=y<rprime|'>>.

  <\proof>
    \ \ introv. revert_last.

    \ \ intros y1 H H0. revert H0. revert y1.

    \ \ induction H;\ 

    \ \ \ \ (intros y2 D2; inversion D2; repeat progress ( eauto \|\|
    econstructor \|\| gen_subst \|\| crush )).
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|fetch_unique> :
  <with|font-shape|italic|gen_subst_db>.

  \;

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation> "<math|n>
  <protectlr|<math|->> <math|\<cal-I\>> <protectlr|(> <math|l> <protectlr|)>
  <protectlr|=> <math|g>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|compute> :
  <with|font-shape|italic|list nat> <math|\<rightarrow\>>
  <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \;

  \| <with|font-shape|italic|Compute<math|<rsub|1>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|n> <math|l> <math|g> <math|k>,

  <\equation*>
    <dfrac|\<cal-I\><protectl|(>l<protectr|)><protectlr|=>
    k<value|sep>n<protectlr|<math|->>k <protectlr|=> g|<math|n>
    <protectlr|<math|->> \<cal-I\><protectlr|<left|(>>l<protectlr|<right|)>><protectlr|=>
    g>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|n> <protectlr|<math|->>
  <math|\<cal-I\>> <protectlr|(> <math|l> <protectlr|)> <protectlr|=>
  <math|g>" := (<with|font-shape|italic|compute> <math|\<cal-I\>> <math|n>
  <math|l> <math|g>).

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|compute_unique> :
  <math|\<forall\>> <math|\<cal-I\>> <math|n> <math|l> <math|y>
  <math|y<rprime|'>>, <math|<math|n> <protectlr|<math|->>
  \<cal-I\><protectlr|<left|(>>l<protectlr|<right|)>><protectlr|=> y>
  <math|\<rightarrow\>> <math|<math|n> <protectlr|<math|->>
  \<cal-I\><protectlr|<left|(>>l<protectlr|<right|)>><protectlr|=>
  y<rprime|'>> <math|\<rightarrow\>> <math|y=y<rprime|'>>.

  <\proof>
    \ \ introv. revert_last.

    \ \ intros y1 H H0. revert H0. revert y1.

    \ \ induction H;\ 

    \ \ \ \ (intros y2 D2; inversion D2; repeat progress ( eauto \|\|
    econstructor \|\| gen_subst \|\| crush )).
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|compute_unique> :
  <with|font-shape|italic|gen_subst_db>.

  \;

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation> "<math|n>
  <protectlr|<math|\<in\>>> <math|\<cal-I\>>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|member> :
  <with|font-shape|italic|list nat> <math|\<rightarrow\>>
  <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \;

  \| <with|font-shape|italic|Member<math|<rsub|1>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|n>,

  <\equation*>
    n<protectlr|<math|\<in\>>><around*|(|n::\<cal-I\>|)>
  </equation*>

  \;

  \| <with|font-shape|italic|Member<math|<rsub|2>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|n> <math|n<rprime|'>>,

  <\equation*>
    <dfrac|n<protectlr|<math|\<in\>>>\<cal-I\>|n<protectlr|<math|\<in\>>><around*|(|n<rprime|'>::\<cal-I\>|)>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|n> <protectlr|<math|\<in\>>>
  <math|\<cal-I\>>" := (<with|font-shape|italic|member> <math|\<cal-I\>>
  <math|n>).

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|domain> :
  <math|\<forall\>> <math|\<cal-I\>> <math|k>,
  <math|k<protectlr|<math|\<in\>>>\<cal-I\>\<rightarrow\>\<exists\>n,\<cal-I\><protectl|(>n<protectr|)><protectlr|=>k>.

  <\proof>
    intros <math|\<cal-I\>> k D.

    induction D.

    - exists 0.

    \ \ apply Fetch<math|<rsub|1>>.

    - destruct IHD.

    \ \ exists (S x).

    \ \ apply Fetch<math|<rsub|2>>.

    \ \ exact H.
  </proof>

  <\comment>
    <subsection|Safe <math|\<lambda\>><rsub|ct>-terms>
  </comment>

  <with|font-series|bold|Inductive> <with|font-shape|italic|term> :
  <with|font-shape|italic|Type> :=

  \| <with|font-shape|italic|Var> (<math|n> : <with|font-shape|italic|nat>)

  \| <with|font-shape|italic|App> (<math|t<rsub|1>> :
  <with|font-shape|italic|term>) (<math|t<rsub|2>> :
  <with|font-shape|italic|term>)

  \| <with|font-shape|italic|Lambda> (<math|t> :
  <with|font-shape|italic|term>)

  \| <with|font-shape|italic|Catch> (<math|t> :
  <with|font-shape|italic|term>)

  \| <with|font-shape|italic|Throw> (<math|\<alpha\>> :
  <with|font-shape|italic|nat>) (<math|t> : <with|font-shape|italic|term>).

  \;

  <with|font-series|bold|Notation> "'<math|<lprime|`>>' <math|x>
  '<math|<lprime|`>>'" := (<with|font-shape|italic|Var> <math|x>) (at level
  80).

  <with|font-series|bold|Notation> "<math|t<rsub|1>> '<protect_lr|<math|*>>'
  <math|t<rsub|2>>" := (<with|font-shape|italic|App> <math|t<rsub|1>>
  <math|t<rsub|2>>) (at level 80).

  <with|font-series|bold|Notation> "'<protect_lr|<math|\<lambda\>>>'
  <math|t>" := (<with|font-shape|italic|Lambda> <math|t>) (at level 80).

  \;

  <with|font-series|bold|Notation> "'<with|font-series|bold|catch>'" :=
  <with|font-shape|italic|Catch> (at level 80).

  <with|font-series|bold|Notation> "'<with|font-series|bold|throw>'" :=
  <with|font-shape|italic|Throw> (at level 80).

  <with|font-series|bold|Notation> "'<with|font-series|bold|get-context>'" :=
  <with|font-shape|italic|Catch> (at level 80).

  <with|font-series|bold|Notation> "'<with|font-series|bold|set-context>'" :=
  <with|font-shape|italic|Throw> (at level 80).

  \;

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  \ "'<with|font-shape|italic|Safe>_' '(' <math|n> ')^(' <math|\<cal-I\>> ,
  <math|\<cal-I\><rsub|\<mu\>>> ')' ( <math|t> )" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|safe> :
  <with|font-shape|italic|term> <math|\<rightarrow\>>
  <with|font-shape|italic|list nat> <math|\<rightarrow\>>
  <with|font-shape|italic|list <math|(>list nat<math|<right|)>>>
  <math|\<rightarrow\>> <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \;

  \| <with|font-shape|italic|Safe<math|<rsub|1>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|n> <math|g> <math|k>,

  <\equation*>
    <dfrac|n<protectlr|<math|->>g <protectlr|=>
    k<value|sep>k<protectlr|<math|\<in\>>>\<cal-I\>|<math-it|Safe><rsub|
    n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|<lprime|`>g<lprime|`>|)>>
  </equation*>

  \;

  \| <with|font-shape|italic|Safe<math|<rsub|2>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|n> <math|t> <math|u>,

  <\equation*>
    <dfrac|<math-it|Safe><rsub| n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|t|)><value|sep><math-it|Safe><rsub|
    n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|u|)>|<math-it|Safe><rsub|
    n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|t<protectlr|<math|*>>u|)>>
  </equation*>

  \;

  \| <with|font-shape|italic|Safe<math|<rsub|3>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|n> <math|t>,

  <\equation*>
    <dfrac|<math-it|Safe><rsub| S n><rsup| <around*|(|S
    n::\<cal-I\>|)>,\<cal-I\><rsub|\<mu\>>><around*|(|t|)>|<math-it|Safe><rsub|
    n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|<protectlr|<math|\<lambda\>>>t|)>>
  </equation*>

  \;

  \| <with|font-shape|italic|Safe<math|<rsub|4>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|n> <math|t>,

  <\equation*>
    <dfrac|<math-it|Safe><rsub| n><rsup| \<cal-I\>,<around*|(|\<cal-I\>::\<cal-I\><rsub|\<mu\>>|)>><around*|(|t|)>|<math-it|Safe><rsub|
    n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|<math-bf|catch><space|1spc>t|)>>
  </equation*>

  \;

  \| <with|font-shape|italic|Safe<math|<rsub|5>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|\<cal-I\><rprime|'>> <math|\<cal-I\><rsub|\<mu\>>>
  <math|n> <math|t> <math|\<alpha\>>,

  <\equation*>
    <dfrac|\<cal-I\><rsub|\<mu\>><protectl|(>\<alpha\><protectr|)>
    <protectlr|=> \<cal-I\><rprime|'><value|sep><math-it|Safe><rsub| n><rsup|
    \<cal-I\><rprime|'>,\<cal-I\><rsub|\<mu\>>><around*|(|t|)>|<math-it|Safe><rsub|
    n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|<math-bf|throw><space|1spc>\<alpha\><space|1spc>t|)>>
  </equation*>

  \;

  <with|font-series|bold|where> "'<with|font-shape|italic|Safe>_' '('
  <math|n> ')^(' <math|\<cal-I\>> , <math|\<cal-I\><rsub|\<mu\>>> ')' (
  <math|t> )" := (<with|font-shape|italic|safe> <math|t> <math|\<cal-I\>>
  <math|\<cal-I\><rsub|\<mu\>>> <math|n>).

  \;

  <\comment>
    <subsection|From local indices to global indices>
  </comment>

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "'<math|\<downarrow\>>_' '(' <math|n> ')^(' <math|\<cal-I\>> ,
  <math|\<cal-I\>\<mu\>> ')' ( <math|t<rsub|1>> ) = <math|t<rsub|2>>" (at
  level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|litogi> :
  <with|font-shape|italic|term> <math|\<rightarrow\>>
  <with|font-shape|italic|list nat> <math|\<rightarrow\>>
  <with|font-shape|italic|list <math|(>list nat<math|<right|)>>>
  <math|\<rightarrow\>> <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|term> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \;

  \| <with|font-shape|italic|Litogi<math|<rsub|1>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|n> <math|g> <math|l>,

  <\equation*>
    <dfrac|<math|n> <protectlr|<math|->> \<cal-I\><protectlr|<left|(>>l<protectlr|<right|)>><protectlr|=>
    g|\<downarrow\><rsub| n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|<lprime|`>l<lprime|`>|)>=<around*|(|<lprime|`>g<lprime|`>|)>>
  </equation*>

  \;

  \| <with|font-shape|italic|Litogi<math|<rsub|2>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|n> <math|t> <math|u>
  <math|t<rprime|'>> <math|u<rprime|'>>,

  <\equation*>
    <dfrac|\<downarrow\><rsub| n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|t|)>=t<rprime|'><value|sep>\<downarrow\><rsub|
    n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|u|)>=u<rprime|'>|\<downarrow\><rsub|
    n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|t<protectlr|<math|*>>u|)>=<around*|(|t<rprime|'><protectlr|<math|*>>u<rprime|'>|)>>
  </equation*>

  \;

  \| <with|font-shape|italic|Litogi<math|<rsub|3>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|n> <math|t>
  <math|t<rprime|'>>,

  <\equation*>
    <dfrac|\<downarrow\><rsub| S n><rsup| <around*|(|S
    n::\<cal-I\>|)>,\<cal-I\><rsub|\<mu\>>><around*|(|t|)>=t<rprime|'>|\<downarrow\><rsub|
    n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|<protectlr|<math|\<lambda\>>>t|)>=<around*|(|<protectlr|<math|\<lambda\>>>t<rprime|'>|)>>
  </equation*>

  \;

  \| <with|font-shape|italic|Litogi<math|<rsub|4>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|n> <math|t>
  <math|t<rprime|'>>,

  <\equation*>
    <dfrac|\<downarrow\><rsub| n><rsup| \<cal-I\>,<around*|(|\<cal-I\>::\<cal-I\><rsub|\<mu\>>|)>><around*|(|t|)>=t<rprime|'>|\<downarrow\><rsub|
    n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|<math-bf|get-context><space|1spc>t|)>=<around*|(|<math-bf|catch><space|1spc>t<rprime|'>|)>>
  </equation*>

  \;

  \| <with|font-shape|italic|Litogi<math|<rsub|5>>> : <math|\<forall\>>
  <math|\<cal-I\>> <math|\<cal-I\><rprime|'>> <math|\<cal-I\><rsub|\<mu\>>>
  <math|n> <math|t> <math|t<rprime|'>> <math|\<alpha\>>,

  <\equation*>
    <dfrac|\<cal-I\><rsub|\<mu\>><protectl|(>\<alpha\><protectr|)>
    <protectlr|=> \<cal-I\><rprime|'><value|sep>\<downarrow\><rsub| n><rsup|
    \<cal-I\><rprime|'>,\<cal-I\><rsub|\<mu\>>><around*|(|t|)>=t<rprime|'>|\<downarrow\><rsub|
    n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|<math-bf|set-context><space|1spc>\<alpha\><space|1spc>t|)>=<around*|(|<math-bf|throw><space|1spc>\<alpha\><space|1spc>t<rprime|'>|)>>
  </equation*>

  \;

  \;

  <with|font-series|bold|where> "'<math|\<downarrow\>>_' '(' <math|n> ')^('
  <math|\<cal-I\>> , <math|\<cal-I\><rsub|\<mu\>>> ')' ( <math|t<rsub|1>> ) =
  <math|t<rsub|2>>" := (<with|font-shape|italic|litogi> <math|t<rsub|1>>
  <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|n> <math|t<rsub|2>>).

  \;

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|litogi_unique> :
  <math|\<forall\>> <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|n>
  <math|x> <math|y> <math|y<rprime|'>>, <math|\<downarrow\><rsub| n><rsup|
  \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|x|)>=y> <math|\<rightarrow\>>
  <math|\<downarrow\><rsub| n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|x|)>=y<rprime|'>>
  <math|\<rightarrow\>> <math|y=y<rprime|'>>.

  <\proof>
    \ \ introv. revert_last.

    \ \ intros y1 H H0. revert H0. revert y1.

    \ \ induction H;\ 

    \ \ \ \ (intros y2 D2; inversion D2; repeat progress ( eauto \|\|
    econstructor \|\| gen_subst \|\| crush )).
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|litogi_unique> :
  <with|font-shape|italic|gen_subst_db>.

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|safe_image> :
  \ <math|\<forall\>> <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|n>
  <math|t<rprime|'>>, <math|<math-it|Safe><rsub| n><rsup|
  \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|t<rprime|'>|)>\<rightarrow\>\<exists\>t,\<downarrow\><rsub|
  n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|t|)>=t<rprime|'>>.

  <\proof>
    \ \ intros <math|\<cal-I\>> <math|\<cal-I\>_\<mu\>> n t D.

    \ \ induction D.

    \ \ - apply domain in H0.

    \ \ \ \ apply minus_swap in H.

    \ \ \ \ destruct H0 as [ l D111 ].

    \ \ \ \ exists (`l`).

    \ \ \ \ apply Litogi_1.

    \ \ \ \ eapply Compute_1.

    \ \ \ \ exact D111.

    \ \ \ \ exact H.

    \ \ - destruct IHD1 as [ t' D11 ].

    \ \ \ \ destruct IHD2 as [ u' D12 ].

    \ \ \ \ exists (t' _*_ u').

    \ \ \ \ apply Litogi_2.

    \ \ \ \ exact D11.

    \ \ \ \ exact D12.

    \ \ - destruct IHD as [ t' D11 ].

    \ \ \ \ exists <math|(_\<lambda\>_> t').

    \ \ \ \ apply Litogi_3.

    \ \ \ \ exact D11.

    \ \ - destruct IHD as [ t' D11 ].

    \ \ \ \ exists (get-context t').

    \ \ \ \ apply Litogi_4.

    \ \ \ \ exact D11.

    \ \ - destruct IHD as [ t' D11 ].

    \ \ \ \ exists (set-context <math|\<alpha\>> t').

    \ \ \ \ eapply Litogi_5.

    \ \ \ \ exact H.

    \ \ \ \ exact D11.
  </proof>

  <\comment>
    <subsection|Abstract machine for <math|\<lambda\><rsub|<math-bf|ct>>>-terms>
  </comment>

  <\comment>
    <subsubsection|Closure, environment, stack and state>
  </comment>

  <with|font-series|bold|Inductive> <with|font-shape|italic|clos> :=

  \| <with|font-shape|italic|Cl> (<math|t> : <with|font-shape|italic|term>)
  (<math|\<cal-E\>> : <with|font-shape|italic|list>
  <with|font-shape|italic|clos>) (<math|\<cal-E\><rsub|\<mu\>>> :
  <with|font-shape|italic|list> (<with|font-shape|italic|list>
  <with|font-shape|italic|clos>)).

  \;

  <with|font-series|bold|Definition> <with|font-shape|italic|stack> :=
  <with|font-shape|italic|list> <with|font-shape|italic|clos>.

  <with|font-series|bold|Definition> <with|font-shape|italic|c_env> :=
  <with|font-shape|italic|list> <with|font-shape|italic|clos>.

  <with|font-series|bold|Definition> <with|font-shape|italic|k_env> :=
  <with|font-shape|italic|list> <with|font-shape|italic|stack>.

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|state> :=

  \| <with|font-shape|italic|St> (<math|t> : <with|font-shape|italic|term>)
  (<math|\<cal-E\>> : <with|font-shape|italic|c_env>)
  (<math|\<cal-E\><rsub|\<mu\>>> : <with|font-shape|italic|k_env>)
  (<math|\<cal-S\>> : <with|font-shape|italic|stack>).

  \;

  <with|font-series|bold|Notation> "<math|<left|\<langle\>>> <math|t> ,
  <math|\<cal-E\>> , <math|\<cal-E\><rsub|\<mu\>>> , <math|\<cal-S\>>
  <math|<left|\<rangle\>>>" := (<with|font-shape|italic|St> <math|t>
  <math|\<cal-E\>> <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-S\>>) (at level
  80).

  <with|font-series|bold|Notation> "[ <math|t> , <math|\<cal-E\>> ,
  <math|\<cal-E\><rsub|\<mu\>>> ]" := (<with|font-shape|italic|Cl> <math|t>
  <math|\<cal-E\>> <math|\<cal-E\><rsub|\<mu\>>>) (at level 80).

  <\comment>
    <subsubsection|Evaluation rules>
  </comment>

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<sigma\><rsub|1>> <math|\<rightsquigarrow\>>
  <math|\<sigma\><rsub|2>>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|step> :
  <with|font-shape|italic|state> <math|\<rightarrow\>>
  <with|font-shape|italic|state> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \;

  \| <with|font-shape|italic|K_var> : <math|\<forall\>> <math|k> <math|t>
  <math|\<cal-E\>> <math|\<cal-E\><rprime|'>> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>> <math|\<cal-S\>>,

  <\equation*>
    <dfrac|\<cal-E\><protectl|(>k<protectr|)> <protectlr|=>
    <around*|[|t,\<cal-E\><rprime|'>,\<cal-E\><rsub|\<mu\>><rprime|'>|]>|<left|\<langle\>>
    <lprime|`>k<lprime|`> , \<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\><left|\<rangle\>>
    \<rightsquigarrow\> <left|\<langle\>>t,\<cal-E\><rprime|'>,\<cal-E\><rsub|\<mu\>><rprime|'>,\<cal-S\><left|\<rangle\>>>
  </equation*>

  \;

  \| <with|font-shape|italic|K_app> : <math|\<forall\>> <math|t> <math|u>
  <math|\<cal-E\>> <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-S\>>,

  <\equation*>
    <left|\<langle\>><around*|(|t<protectlr|<math|*>>u|)>,
    \<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\><left|\<rangle\>>
    \<rightsquigarrow\> <left|\<langle\>>t,\<cal-E\>,\<cal-E\><rsub|\<mu\>>,<around*|[|u,\<cal-E\>,\<cal-E\><rsub|\<mu\>>|]>::\<cal-S\><left|\<rangle\>>
  </equation*>

  \;

  \| <with|font-shape|italic|K_abs> : <math|\<forall\>> <math|t> <math|c>
  <math|\<cal-E\>> <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-S\>>,

  <\equation*>
    <left|\<langle\>><protectlr|<math|\<lambda\>>>t,
    \<cal-E\>,\<cal-E\><rsub|\<mu\>>,c ::\<cal-S\><left|\<rangle\>>
    \<rightsquigarrow\> <left|\<langle\>>t,<around*|(|c
    ::\<cal-E\>|)>,\<cal-E\><rsub|\<mu\>>,\<cal-S\><left|\<rangle\>>
  </equation*>

  \;

  \| <with|font-shape|italic|K_catch> : <math|\<forall\>> <math|t>
  <math|\<cal-E\>> <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-S\>>,

  <\equation*>
    <left|\<langle\>><math-bf|catch><space|1spc>t,
    \<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\><left|\<rangle\>>
    \<rightsquigarrow\> <left|\<langle\>>t,\<cal-E\>,<around*|(|\<cal-S\>
    ::\<cal-E\><rsub|\<mu\>>|)>,\<cal-S\><left|\<rangle\>>
  </equation*>

  \;

  \| <with|font-shape|italic|K_throw> : <math|\<forall\>> <math|t>
  <math|\<alpha\>> <math|\<cal-E\>> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-S\>> <math|\<cal-S\><rprime|'>>,

  <\equation*>
    <dfrac|\<cal-E\><rsub|\<mu\>><protectl|(>\<alpha\><protectr|)>
    <protectlr|=> \<cal-S\><rprime|'>|<left|\<langle\>><math-bf|throw><space|1spc>\<alpha\><space|1spc>t,
    \<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\><left|\<rangle\>>
    \<rightsquigarrow\> <left|\<langle\>>t,\<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\><rprime|'><left|\<rangle\>>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<sigma\><rsub|1>>
  <math|\<rightsquigarrow\>> <math|\<sigma\><rsub|2>>" :=
  (<with|font-shape|italic|step> <math|\<sigma\><rsub|1>>
  <math|\<sigma\><rsub|2>>).

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|step_unique> :
  <math|\<forall\>> <math|\<sigma\><rsub|1>> <math|\<sigma\><rsub|2>>
  <math|\<sigma\><rsub|2><rprime|'>>, <math|\<sigma\><rsub|1>>
  <math|\<rightsquigarrow\>> <math|\<sigma\><rsub|2>> <math|\<rightarrow\>>
  <math|\<sigma\><rsub|1>> <math|\<rightsquigarrow\>>
  <math|\<sigma\><rsub|2><rprime|'>> <math|\<rightarrow\>>
  <math|\<sigma\><rsub|2>=\<sigma\><rsub|2><rprime|'>>.

  <\proof>
    \ \ introv. revert_last.

    \ \ intros y1 H H0. revert H0. revert y1.

    \ \ induction H;\ 

    \ \ \ \ (intros y2 D2; inversion D2; repeat progress ( eauto \|\|
    econstructor \|\| gen_subst \|\| crush )).
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|step_unique> : <with|font-shape|italic|gen_subst_db>.

  \;

  <\comment>
    <subsection|Abstract machine for safe
    <math|\<lambda\><rsub|<math-bf|ct>>>-terms (with local environments)>
  </comment>

  <\comment>
    <subsubsection|Closure, environment, stack and state>
  </comment>

  <with|font-series|bold|Inductive> <with|font-shape|italic|clos_l> :=

  \| <with|font-shape|italic|Cl_l> (<math|t> : <with|font-shape|italic|term>)
  (<math|\<cal-L\>> : <with|font-shape|italic|list>
  <with|font-shape|italic|clos_l>) (<math|\<cal-L\><rsub|\<mu\>>> :
  <with|font-shape|italic|list> (<with|font-shape|italic|list>
  <with|font-shape|italic|clos_l>)) (<math|\<cal-E\><rsub|\<mu\>>> :
  <with|font-shape|italic|list> (<with|font-shape|italic|list>
  <with|font-shape|italic|clos_l>)).

  \;

  <with|font-series|bold|Definition> <with|font-shape|italic|stack_l> :=
  <with|font-shape|italic|list> <with|font-shape|italic|clos_l>.

  <with|font-series|bold|Definition> <with|font-shape|italic|l_env_l> :=
  <with|font-shape|italic|list> <with|font-shape|italic|clos_l>.

  <with|font-series|bold|Definition> <with|font-shape|italic|m_env_l> :=
  <with|font-shape|italic|list> <with|font-shape|italic|l_env_l>.

  <with|font-series|bold|Definition> <with|font-shape|italic|k_env_l> :=
  <with|font-shape|italic|list> <with|font-shape|italic|stack_l>.

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|state_l> :=

  \| <with|font-shape|italic|St_l> (<math|t> : <with|font-shape|italic|term>)
  (<math|\<cal-L\>> : <with|font-shape|italic|l_env_l>)
  (<math|\<cal-L\><rsub|\<mu\>>> : <with|font-shape|italic|m_env_l>)
  (<math|\<cal-E\><rsub|\<mu\>>> : <with|font-shape|italic|k_env_l>)
  (<math|\<cal-S\>> : <with|font-shape|italic|stack_l>).

  \;

  <with|font-series|bold|Notation> "<math|<left|\<langle\>>> <math|t> ,
  <math|\<cal-L\>> , <math|\<cal-L\><rsub|\<mu\>>> ,
  <math|\<cal-E\><rsub|\<mu\>>> , <math|\<cal-S\>> <math|<left|\<rangle\>>>"
  := (<with|font-shape|italic|St_l> <math|t> <math|\<cal-L\>>
  <math|\<cal-L\><rsub|\<mu\>>> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-S\>>) (at level 80).

  <with|font-series|bold|Notation> "[ <math|t> , <math|\<cal-L\>> ,
  <math|\<cal-L\><rsub|\<mu\>>> , <math|\<cal-E\><rsub|\<mu\>>> ]" :=
  (<with|font-shape|italic|Cl_l> <math|t> <math|\<cal-L\>>
  <math|\<cal-L\><rsub|\<mu\>>> <math|\<cal-E\><rsub|\<mu\>>>) (at level 80).

  <\comment>
    <subsubsection|Evaluation rules>
  </comment>

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<sigma\><rsub|1>> <math|\<rightsquigarrow\>> '^<math|l>'
  <math|\<sigma\><rsub|2>>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|step_l> :
  <with|font-shape|italic|state_l> <math|\<rightarrow\>>
  <with|font-shape|italic|state_l> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \;

  \| <with|font-shape|italic|L_var> : <math|\<forall\>> <math|k> <math|t>
  <math|\<cal-L\>> <math|\<cal-L\><rprime|'>> <math|\<cal-L\><rsub|\<mu\>>>
  <math|\<cal-L\><rsub|\<mu\>><rprime|'>> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>> <math|\<cal-S\>>,

  <\equation*>
    <dfrac|\<cal-L\><protectl|(>k<protectr|)> <protectlr|=>
    <around*|[|t,\<cal-L\><rprime|'>,\<cal-L\><rsub|\<mu\>><rprime|'>,\<cal-E\><rsub|\<mu\>><rprime|'>|]>|\<langle\>
    <lprime|`>k<lprime|`> ,\<cal-L\>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>>,\<cal-S\>\<rangle\>
    \<rightsquigarrow\><rsup|l>\<langle\> t,
    \<cal-L\><rprime|'>,\<cal-L\><rsub|\<mu\>><rprime|'>,\<cal-E\><rsub|\<mu\>><rprime|'>,\<cal-S\>\<rangle\>>
  </equation*>

  \;

  \| <with|font-shape|italic|L_app> : <math|\<forall\>> <math|t> <math|u>
  <math|\<cal-L\>> <math|\<cal-L\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-S\>>,

  <\equation*>
    \<langle\><around*|(|t<protectlr|<math|*>>u|)>,\<cal-L\>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>>,\<cal-S\>\<rangle\>
    \<rightsquigarrow\><rsup|l>\<langle\>
    t,\<cal-L\>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>>,<around*|[|u,\<cal-L\>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>>|]>::\<cal-S\>\<rangle\>
  </equation*>

  \;

  \| <with|font-shape|italic|L_abs> : <math|\<forall\>> <math|t> <math|c>
  <math|\<cal-L\>> <math|\<cal-L\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>>> <math|<with|math-font|cal|S<rprime|'>>>,

  <\equation*>
    <around|\<langle\>|<protectlr|<math|\<lambda\>>>t,
    \<cal-L\>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>>,c
    ::<with|math-font|cal|S<rprime|'>>|\<rangle\>>
    \<rightsquigarrow\><rsup|l><around*|\<langle\>| t, <around*|(|c ::
    \<cal-L\>|)>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>>,<with|math-font|cal|S<rprime|'>>|\<rangle\>>
  </equation*>

  \;

  \| <with|font-shape|italic|L_catch> : <math|\<forall\>> <math|t>
  <math|\<cal-L\>> <math|\<cal-L\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-S\>>,

  <\equation*>
    <around|\<langle\>|<math-bf|get-context><space|1spc>t,
    \<cal-L\>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>>,\<cal-S\>|\<rangle\>>
    \<rightsquigarrow\><rsup|l><around*|\<langle\>| t,
    \ \<cal-L\>,<around*|(|\<cal-L\> ::\<cal-L\><rsub|\<mu\>>|)>,<around*|(|\<cal-S\>
    ::\<cal-E\><rsub|\<mu\>>|)>,\<cal-S\>|\<rangle\>>
  </equation*>

  \;

  \| <with|font-shape|italic|L_throw> : <math|\<forall\>> <math|t>
  <math|\<alpha\>> <math|\<cal-L\>> <math|\<cal-L\><rprime|'>>
  <math|\<cal-L\><rsub|\<mu\>>> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-S\>> <math|\<cal-S\><rprime|'>>,

  <\equation*>
    <dfrac|\<cal-L\><rsub|\<mu\>><protectl|(>\<alpha\><protectr|)>
    <protectlr|=> <with|math-font|cal|\<cal-L\><rprime|'>><value|sep>\<cal-E\><rsub|\<mu\>><protectl|(>\<alpha\><protectr|)>
    <protectlr|=> \<cal-S\><rprime|'>|<around|\<langle\>|<math-bf|set-context><space|1spc>\<alpha\><space|1spc>t,\<cal-L\>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>>,\<cal-S\>|\<rangle\>>
    \<rightsquigarrow\><rsup|l><around*|\<langle\>| t,
    \ \<cal-L\><rprime|'>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>>,\<cal-S\><rprime|'>|\<rangle\>>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<sigma\><rsub|1>>
  <math|\<rightsquigarrow\>> '^<math|l>' <math|\<sigma\><rsub|2>>" :=
  (<with|font-shape|italic|step_l> <math|\<sigma\><rsub|1>>
  <math|\<sigma\><rsub|2>>).

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|step_l_unique> :
  <math|\<forall\>> <math|\<sigma\><rsub|1>> <math|\<sigma\><rsub|2>>
  <math|\<sigma\><rsub|2><rprime|'>>, <math|\<sigma\><rsub|1>>
  <math|\<rightsquigarrow\><rsup|l>> <math|\<sigma\><rsub|2>>
  <math|\<rightarrow\>> <math|\<sigma\><rsub|1>>
  <math|\<rightsquigarrow\><rsup|l>> <math|\<sigma\><rsub|2><rprime|'>>
  <math|\<rightarrow\>> <math|\<sigma\><rsub|2>=\<sigma\><rsub|2><rprime|'>>.

  <\proof>
    \ \ introv. revert_last.

    \ \ intros y1 H H0. revert H0. revert y1.

    \ \ induction H;\ 

    \ \ \ \ (intros y2 D2; inversion D2; repeat progress ( eauto \|\|
    econstructor \|\| gen_subst \|\| crush )).
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|step_l_unique> :
  <with|font-shape|italic|gen_subst_db>.

  \;

  \;

  <\comment>
    <section|Bisimulations>
  </comment>

  <\comment>
    <subsection|Abstract machine for safe
    <math|\<lambda\><rsub|<math-bf|ct>>>-terms (with indirection tables)>
  </comment>

  <\comment>
    <subsubsection|Closure, environment, stack and state>
  </comment>

  <with|font-series|bold|Inductive> <with|font-shape|italic|clos_i> :=

  \| <with|font-shape|italic|Cl_i> (<math|t> : <with|font-shape|italic|term>)
  (<math|n> : <with|font-shape|italic|nat>) (<math|\<cal-I\>> :
  <with|font-shape|italic|vector>) (<math|\<cal-I\><rsub|\<mu\>>> :
  <with|font-shape|italic|table>) (<math|\<cal-E\>> :
  <with|font-shape|italic|list> <with|font-shape|italic|clos_i>)
  (<math|\<cal-E\><rsub|\<mu\>>> : <with|font-shape|italic|list>
  (<with|font-shape|italic|list> <with|font-shape|italic|clos_i>)).

  \;

  <with|font-series|bold|Definition> <with|font-shape|italic|stack_i> :=
  <with|font-shape|italic|list> <with|font-shape|italic|clos_i>.

  <with|font-series|bold|Definition> <with|font-shape|italic|c_env_i> :=
  <with|font-shape|italic|list> <with|font-shape|italic|clos_i>.

  <with|font-series|bold|Definition> <with|font-shape|italic|k_env_i> :=
  <with|font-shape|italic|list> <with|font-shape|italic|stack_i>.

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|state_i> :=

  \| <with|font-shape|italic|St_i> (<math|t> : <with|font-shape|italic|term>)
  (<math|n> : <with|font-shape|italic|nat>) (<math|\<cal-I\>> :
  <with|font-shape|italic|vector>) (<math|\<cal-I\><rsub|\<mu\>>> :
  <with|font-shape|italic|table>) (<math|\<cal-E\>> :
  <with|font-shape|italic|c_env_i>) (<math|\<cal-E\><rsub|\<mu\>>> :
  <with|font-shape|italic|k_env_i>) (<math|\<cal-S\>> :
  <with|font-shape|italic|stack_i>).

  \;

  <with|font-series|bold|Notation> "<math|<left|\<langle\>>> <math|t> ,
  <math|n> , <math|\<cal-I\>> , <math|\<cal-I\><rsub|\<mu\>>> ,
  <math|\<cal-E\>> , <math|\<cal-E\><rsub|\<mu\>>> , <math|\<cal-S\>>
  <math|<left|\<rangle\>>>" := (<with|font-shape|italic|St_i> <math|t>
  <math|n> <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|\<cal-E\>>
  <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-S\>>) (at level 80).

  <with|font-series|bold|Notation> "[ <math|t> , <math|n> , <math|\<cal-I\>>
  , <math|\<cal-I\><rsub|\<mu\>>> , <math|\<cal-E\>> ,
  <math|\<cal-E\><rsub|\<mu\>>> ]" := (<with|font-shape|italic|Cl_i> <math|t>
  <math|n> <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|\<cal-E\>>
  <math|\<cal-E\><rsub|\<mu\>>>) (at level 80).

  <\comment>
    <subsubsection|Evaluation rules>
  </comment>

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<sigma\><rsub|1>> <math|\<rightsquigarrow\>> '^<math|i>'
  <math|\<sigma\><rsub|2>>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|step_i> :
  <with|font-shape|italic|state_i> <math|\<rightarrow\>>
  <with|font-shape|italic|state_i> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \;

  \| <with|font-shape|italic|I_var> : <math|\<forall\>> <math|n>
  <math|n<rprime|'>> <math|l> <math|g> <math|t> <math|\<cal-I\>>
  <math|\<cal-I\><rprime|'>> <math|\<cal-I\><rsub|\<mu\>>>
  <math|\<cal-I\><rsub|\<mu\>><rprime|'>> <math|\<cal-E\>>
  <math|\<cal-E\><rprime|'>> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>> <math|\<cal-S\>>,

  <\equation*>
    <dfrac|<math|n> <protectlr|<math|->> \<cal-I\><protectlr|<left|(>>l<protectlr|<right|)>>
    <protectlr|=> g<value|sep>\<cal-E\><protectl|(>g<protectr|)>
    <protectlr|=> <around*|[|t,n<rprime|'>,
    \<cal-I\><rprime|'>,\<cal-I\><rsub|\<mu\>><rprime|'>,\<cal-E\><rprime|'>,\<cal-E\><rsub|\<mu\>><rprime|'>|]>|\<langle\>
    <lprime|`>l<lprime|`> ,n, \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\>\<rangle\>
    \<rightsquigarrow\><rsup|i>\<langle\> t,n<rprime|'>,
    \<cal-I\><rprime|'>,\<cal-I\><rsub|\<mu\>><rprime|'>,\<cal-E\><rprime|'>,\<cal-E\><rsub|\<mu\>><rprime|'>,\<cal-S\>\<rangle\>>
  </equation*>

  \;

  \| <with|font-shape|italic|I_app> : <math|\<forall\>> <math|n> <math|t>
  <math|u> <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|\<cal-E\>>
  <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-S\>>,

  <\equation*>
    \<langle\><around*|(|t<protectlr|<math|*>>u|)>,n,
    \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\>\<rangle\>
    \<rightsquigarrow\><rsup|i>\<langle\> t,n,
    \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>,<around*|[|u,n,
    \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>|]>::\<cal-S\>\<rangle\>
  </equation*>

  \;

  \| <with|font-shape|italic|I_abs> : <math|\<forall\>> <math|n> <math|t>
  <math|c> <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|\<cal-E\>>
  <math|\<cal-E\><rsub|\<mu\>>> <math|<with|math-font|cal|S<rprime|'>>>,

  <\equation*>
    <around|\<langle\>|<protectlr|<math|\<lambda\>>>t,n,
    \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>,c
    ::<with|math-font|cal|S<rprime|'>>|\<rangle\>>
    \<rightsquigarrow\><rsup|i><around*|\<langle\>| t, <around*|(|S
    n|)>,<around*|(|<around*|(| S n|)> ::
    \<cal-I\>|)>,\<cal-I\><rsub|\<mu\>>,c
    ::\<cal-E\>,\<cal-E\><rsub|\<mu\>>,<with|math-font|cal|S<rprime|'>>|\<rangle\>>
  </equation*>

  \;

  \| <with|font-shape|italic|I_catch> : <math|\<forall\>> <math|n> <math|t>
  <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|\<cal-E\>>
  <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-S\>>,

  <\equation*>
    <around|\<langle\>|<math-bf|get-context><space|1spc>t,n,
    \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\>|\<rangle\>>
    \<rightsquigarrow\><rsup|i><around*|\<langle\>| t, n,
    \<cal-I\>,<around*|(|\<cal-I\> ::\<cal-I\><rsub|\<mu\>>|)>,\<cal-E\>,<around*|(|\<cal-S\>
    ::\<cal-E\><rsub|\<mu\>>|)>,\<cal-S\>|\<rangle\>>
  </equation*>

  \;

  \| <with|font-shape|italic|I_throw> : <math|\<forall\>> <math|n> <math|t>
  <math|\<alpha\>> <math|\<cal-I\>> <math|\<cal-I\><rprime|'>>
  <math|\<cal-I\><rsub|\<mu\>>> <math|\<cal-E\>>
  <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-S\>> <math|\<cal-S\><rprime|'>>,

  <\equation*>
    <dfrac|\<cal-I\><rsub|\<mu\>><protectl|(>\<alpha\><protectr|)>
    <protectlr|=> <with|math-font|cal|\<cal-I\><rprime|'>><value|sep>\<cal-E\><rsub|\<mu\>><protectl|(>\<alpha\><protectr|)>
    <protectlr|=> \<cal-S\><rprime|'>|<around|\<langle\>|<math-bf|set-context><space|1spc>\<alpha\><space|1spc>t,n,
    \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\>|\<rangle\>>
    \<rightsquigarrow\><rsup|i><around*|\<langle\>| t, n,
    \<cal-I\><rprime|'>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\><rprime|'>|\<rangle\>>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<sigma\><rsub|1>>
  <math|\<rightsquigarrow\>> '^<math|i>' <math|\<sigma\><rsub|2>>" :=
  (<with|font-shape|italic|step_i> <math|\<sigma\><rsub|1>>
  <math|\<sigma\><rsub|2>>).

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|step_i_unique> :
  <math|\<forall\>> <math|\<sigma\><rsub|1>> <math|\<sigma\><rsub|2>>
  <math|\<sigma\><rsub|2><rprime|'>>, <math|\<sigma\><rsub|1>>
  <math|\<rightsquigarrow\><rsup|i>> <math|\<sigma\><rsub|2>>
  <math|\<rightarrow\>> <math|\<sigma\><rsub|1>>
  <math|\<rightsquigarrow\><rsup|i>> <math|\<sigma\><rsub|2><rprime|'>>
  <math|\<rightarrow\>> <math|\<sigma\><rsub|2>=\<sigma\><rsub|2><rprime|'>>.

  <\proof>
    \ \ introv. revert_last.

    \ \ intros y1 H H0. revert H0. revert y1.

    \ \ induction H;\ 

    \ \ \ \ (intros y2 D2; inversion D2; repeat progress ( eauto \|\|
    econstructor \|\| gen_subst \|\| crush )).
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|step_i_unique> :
  <with|font-shape|italic|gen_subst_db>.

  <\comment>
    <subsection|Lock-step simulation <math|<around*|(|-|)><rsup|\<star\>>>>
  </comment>

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|<with|math-font|cal|c>> '^<math|\<star\>>=_<math|c>'
  <math|c<rprime|'>>" (at level 30).

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<cal-S\>> '^<math|\<star\>>=_<math|s>' <math|\<cal-S\><rprime|'>>"
  (at level 30).

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<cal-E\>> '^<math|\<star\>>=_<math|e>' <math|\<cal-E\><rprime|'>>"
  (at level 30).

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<cal-E\><rsub|\<mu\>>> '^<math|\<star\>>=_<math|k>'
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|translate_clos> :
  <with|font-shape|italic|clos_i> <math|\<rightarrow\>>
  <with|font-shape|italic|clos> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \| <with|font-shape|italic|T_cl> : <math|\<forall\>> <math|n> <math|t>
  <math|u> <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|\<cal-E\>>
  <math|\<cal-E\><rprime|'>> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>>,

  <\equation*>
    <dfrac|\<downarrow\><rsub| n><rsup| \<cal-I\>,\<cal-I\><rsub|\<mu\>>><around*|(|t|)>=u<value|sep>\<cal-E\><rsup|\<star\>>=<rsub|e>
    \<cal-E\><rprime|'><value|sep>\<cal-E\><rsub|\<mu\>><rsup|\<star\>>=<rsub|k>
    \<cal-E\><rsub|\<mu\>><rprime|'>|<around*|[|t,n,
    \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>|]><rsup|\<star\>>=<rsub|c>
    <around*|[| u,\<cal-E\><rprime|'>,\<cal-E\><rsub|\<mu\>><rprime|'>|]>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|<with|math-font|cal|c>>
  '^<math|\<star\>>=_<math|c>' <math|c<rprime|'>>" :=
  (<with|font-shape|italic|translate_clos> <math|<with|math-font|cal|c>>
  <math|c<rprime|'>>)

  \;

  <with|font-series|bold|with> <with|font-shape|italic|translate_stack> :
  <with|font-shape|italic|stack_i> <math|\<rightarrow\>>
  <with|font-shape|italic|stack> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \| <with|font-shape|italic|T_stack<math|<rsub|1>>> :

  <\equation*>
    <math-it|nil> <rsup|\<star\>>=<rsub|s> <math-it|nil>
  </equation*>

  \;

  \| <with|font-shape|italic|T_stack<math|<rsub|2>>> : <math|\<forall\>>
  <math|c> <math|c<rprime|'>> <math|\<cal-S\>> <math|\<cal-S\><rprime|'>>,

  <\equation*>
    <dfrac|c<rsup|\<star\>>=<rsub|c> c<rprime|'><value|sep>\<cal-S\><rsup|\<star\>>=<rsub|s>
    \<cal-S\><rprime|'>|<with|math-font|cal|<around*|(|c
    ::\<cal-S\>|)>><rsup|\<star\>>=<rsub|s>
    <around*|(|<with|math-font|cal|c<rprime|'> ::\<cal-S\><rprime|'>>|)>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<cal-S\>>
  '^<math|\<star\>>=_<math|s>' <math|\<cal-S\><rprime|'>>" :=
  (<with|font-shape|italic|translate_stack> <math|\<cal-S\>>
  <math|\<cal-S\><rprime|'>>)

  \;

  <with|font-series|bold|with> <with|font-shape|italic|translate_c_env> :
  <with|font-shape|italic|c_env_i> <math|\<rightarrow\>>
  <with|font-shape|italic|c_env> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \| <with|font-shape|italic|T_c_env<math|<rsub|1>>> :

  <\equation*>
    <math-it|nil> <rsup|\<star\>>=<rsub|e> <math-it|nil>
  </equation*>

  \;

  \| <with|font-shape|italic|T_c_env<math|<rsub|2>>> : <math|\<forall\>>
  <math|c> <math|c<rprime|'>> <math|\<cal-E\>> <math|\<cal-E\><rprime|'>>,

  <\equation*>
    <dfrac|c<rsup|\<star\>>=<rsub|c> c<rprime|'><value|sep>\<cal-E\><rsup|\<star\>>=<rsub|e>
    \<cal-E\><rprime|'>|<with|math-font|cal|<around*|(|c
    ::\<cal-E\>|)>><rsup|\<star\>>=<rsub|e> <around*|(|c<rprime|'>
    ::\<cal-E\><rprime|'>|)>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<cal-E\>>
  '^<math|\<star\>>=_<math|e>' <math|\<cal-E\><rprime|'>>" :=
  (<with|font-shape|italic|translate_c_env> <math|\<cal-E\>>
  <math|\<cal-E\><rprime|'>>)

  \;

  <with|font-series|bold|with> <with|font-shape|italic|translate_k_env> :
  <with|font-shape|italic|k_env_i> <math|\<rightarrow\>>
  <with|font-shape|italic|k_env> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \| <with|font-shape|italic|T_k_env<math|<rsub|1>>> :

  <\equation*>
    <math-it|nil><rsup|\<star\>>=<rsub|k> <math-it|nil>
  </equation*>

  \;

  \| <with|font-shape|italic|T_k_env<math|<rsub|2>>> : <math|\<forall\>>
  <math|\<cal-S\>> <math|\<cal-S\><rprime|'>> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>>,

  <\equation*>
    <dfrac|\<cal-S\><rsup|\<star\>>=<rsub|s>
    \<cal-S\><rprime|'><value|sep>\<cal-E\><rsub|\<mu\>><rsup|\<star\>>=<rsub|k>
    \<cal-E\><rsub|\<mu\>><rprime|'>|<with|math-font|cal|<around*|(|\<cal-S\>
    ::\<cal-E\><rsub|\<mu\>>|)>><rsup|\<star\>>=<rsub|k>
    <around*|(|\<cal-S\><rprime|'> ::\<cal-E\><rsub|\<mu\>><rprime|'>|)>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<cal-E\><rsub|\<mu\>>>
  '^<math|\<star\>>=_<math|k>' <math|\<cal-E\><rsub|\<mu\>><rprime|'>>" :=
  (<with|font-shape|italic|translate_k_env> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>>).

  \;

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<sigma\>> '^<math|\<star\>>=_<math|\<sigma\>>'
  <math|\<sigma\><rprime|'>>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|translate_st> :
  <with|font-shape|italic|state_i> <math|\<rightarrow\>>
  <with|font-shape|italic|state> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \| <with|font-shape|italic|T_st> : <math|\<forall\>> <math|n> <math|t>
  <math|u> <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|\<cal-E\>>
  <math|\<cal-E\><rprime|'>> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>> <math|\<cal-S\>>
  <math|\<cal-S\><rprime|'>>,

  <\equation*>
    <dfrac|<around*|[| t,n, \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>|]><rsup|\<star\>>=<rsub|c>
    <around*|[|u,\<cal-E\><rprime|'>,\<cal-E\><rsub|\<mu\>><rprime|'>|]><value|sep>\<cal-S\><rsup|\<star\>>=<rsub|s>
    \<cal-S\><rprime|'>|\<langle\> t,n, \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\>\<rangle\><rsup|\<star\>>=<rsub|\<sigma\>>
    \<langle\> u,\<cal-E\><rprime|'>,\<cal-E\><rsub|\<mu\>><rprime|'>,\<cal-S\><rprime|'>\<rangle\>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<sigma\>>
  '^<math|\<star\>>=_<math|\<sigma\>>' <math|\<sigma\><rprime|'>>" :=
  (<with|font-shape|italic|translate_st> <math|\<sigma\>>
  <math|\<sigma\><rprime|'>>).

  \;

  <with|font-series|bold|Scheme> <with|font-shape|italic|translate_clos_mut>
  := <with|font-series|bold|Induction> <with|font-series|bold|for>
  <with|font-shape|italic|translate_clos> <with|font-series|bold|Sort>
  <with|font-shape|italic|Prop>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_stack_mut> :=
  <with|font-series|bold|Induction> <with|font-series|bold|for>
  <with|font-shape|italic|translate_stack> <with|font-series|bold|Sort>
  <with|font-shape|italic|Prop>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_c_env_mut> :=
  <with|font-series|bold|Induction> <with|font-series|bold|for>
  <with|font-shape|italic|translate_c_env> <with|font-series|bold|Sort>
  <with|font-shape|italic|Prop>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_k_env_mut> :=
  <with|font-series|bold|Induction> <with|font-series|bold|for>
  <with|font-shape|italic|translate_k_env> <with|font-series|bold|Sort>
  <with|font-shape|italic|Prop>.

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|translate_clos_unique>
  : <math|\<forall\>> <math|c> <math|c<rsub|2>> <math|c<rsub|2><rprime|'>>,
  <math|c<rsup|\<star\>>=<rsub|c> c<rsub|2>> <math|\<rightarrow\>>
  <math|c<rsup|\<star\>>=<rsub|c> c<rsub|2><rprime|'>> <math|\<rightarrow\>>
  <math|c<rsub|2>=c<rsub|2><rprime|'>>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_stack_unique> : <math|\<forall\>>
  <math|s> <math|s<rsub|2>> <math|s<rsub|2><rprime|'>>,
  <math|s<rsup|\<star\>>=<rsub|s> s<rsub|2>> <math|\<rightarrow\>>
  <math|s<rsup|\<star\>>=<rsub|s> s<rsub|2><rprime|'>> <math|\<rightarrow\>>
  <math|s<rsub|2>=s<rsub|2><rprime|'>>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_c_env_unique> : <math|\<forall\>>
  <math|e> <math|e<rsub|2>> <math|e<rsub|2><rprime|'>>,
  <math|e<rsup|\<star\>>=<rsub|e> e<rsub|2>> <math|\<rightarrow\>>
  <math|e<rsup|\<star\>>=<rsub|e> e<rsub|2><rprime|'>> <math|\<rightarrow\>>
  <math|e<rsub|2>=e<rsub|2><rprime|'>>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_k_env_unique> : <math|\<forall\>>
  <math|k> <math|k<rsub|2>> <math|k<rsub|2><rprime|'>>,
  <math|k<rsup|\<star\>>=<rsub|k> k<rsub|2>> <math|\<rightarrow\>>
  <math|k<rsup|\<star\>>=<rsub|k> k<rsub|2><rprime|'>> <math|\<rightarrow\>>
  <math|k<rsub|2>=k<rsub|2><rprime|'>>.

  <\proof>
    \ \ <with|font-series|bold|Tactic> <with|font-series|bold|Notation>
    <verbatim|"><with|font-shape|italic|apply_unique_tac><verbatim|">
    tactic(l) :=

    \ \ \ \ (unique_tac

    \ \ \ \ \ \ \ (l (unique_spec_1 <with|font-shape|italic|translate_clos>)

    \ \ \ \ \ \ \ \ \ \ (unique_spec_1 <with|font-shape|italic|translate_stack>)

    \ \ \ \ \ \ \ \ \ \ (unique_spec_1 <with|font-shape|italic|translate_c_env>)

    \ \ \ \ \ \ \ \ \ \ (unique_spec_1 <with|font-shape|italic|translate_k_env>))).

    \ \ - <with|font-shape|italic|apply_unique_tac>
    <with|font-shape|italic|translate_clos_mut>.

    \ \ - <with|font-shape|italic|apply_unique_tac>
    <with|font-shape|italic|translate_stack_mut>.

    \ \ - <with|font-shape|italic|apply_unique_tac>
    <with|font-shape|italic|translate_c_env_mut>.

    \ \ - <with|font-shape|italic|apply_unique_tac>
    <with|font-shape|italic|translate_k_env_mut>.
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|translate_clos_unique> :
  <with|font-shape|italic|gen_subst_db>.

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|translate_stack_unique> :
  <with|font-shape|italic|gen_subst_db>.

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|translate_c_env_unique> :
  <with|font-shape|italic|gen_subst_db>.

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|translate_k_env_unique> :
  <with|font-shape|italic|gen_subst_db>.

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|translate_st_unique>
  : <math|\<forall\>> <math|\<sigma\>> <math|\<sigma\><rsub|2>>
  <math|\<sigma\><rsub|2><rprime|'>>, <math|\<sigma\><rsup|\<star\>>=<rsub|\<sigma\>>
  \<sigma\><rsub|2>> <math|\<rightarrow\>>
  <math|\<sigma\><rsup|\<star\>>=<rsub|\<sigma\>>
  \<sigma\><rsub|2><rprime|'>> <math|\<rightarrow\>>
  <math|\<sigma\><rsub|2>=\<sigma\><rsub|2><rprime|'>>.

  <\proof>
    \ \ simple_unique_tac.
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|translate_st_unique> :
  <with|font-shape|italic|gen_subst_db>.

  <\comment>
    <subsubsection|Soundness>
  </comment>

  <with|font-series|bold|Lemma> <with|font-shape|italic|fetch_sound> :
  \ <math|\<forall\>> <math|\<cal-E\>> <math|\<cal-E\><rprime|'>>,
  <math|\<cal-E\><rsup|\<star\>>=<rsub|e>
  \<cal-E\><rprime|'>\<rightarrow\>\<forall\><space|1spc>n<space|1spc>c,<space|1spc>\<cal-E\><protectl|(>n<protectr|)>
  <protectlr|=> c\<rightarrow\>\<exists\>c<rprime|'>,c<rsup|\<star\>>=<rsub|c>
  c<rprime|'>\<wedge\>\<cal-E\><rprime|'><protectl|(>n<protectr|)>
  <protectlr|=> c<rprime|'>>.

  <\proof>
    \ \ intros <math|\<cal-E\>> <math|\<cal-E\><rprime|'>> D1 n c D2.

    \ \ revert <math|\<cal-E\><rprime|'>> D1.

    \ \ induction D2.

    \ \ - intros <math|\<cal-E\><rprime|'>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ exists c'.

    \ \ \ \ split.

    \ \ \ \ * assumption.

    \ \ \ \ * apply Fetch_1.

    \ \ - intros <math|\<cal-E\><rprime|'>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ apply IHD2 in H3.

    \ \ \ \ destruct H3.

    \ \ \ \ destruct H3.

    \ \ \ \ exists x.

    \ \ \ \ split. \ \ \ 

    \ \ \ \ * exact H3.

    \ \ \ \ * apply Fetch_2.

    \ \ \ \ \ \ exact H4.\ 
  </proof>

  <with|font-series|bold|Lemma> <with|font-shape|italic|fetch_mu_sound> :
  \ <math|\<forall\>> <math|n> <math|\<cal-S\>>
  <math|\<cal-E\><rsub|\<mu\>>>, <math|\<cal-E\><rsub|\<mu\>><protectl|(>n<protectr|)>
  <protectlr|=> \<cal-S\>\<rightarrow\>\<forall\>\<cal-E\><rprime|'><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>><rsup|\<star\>>=<rsub|k>
  \<cal-E\><rprime|'><rsub|\<mu\>>\<rightarrow\>\<exists\>\<cal-S\><rprime|'>,\<cal-S\><rsup|\<star\>>=<rsub|s>
  \<cal-S\><rprime|'>\<wedge\>\<cal-E\><rprime|'><rsub|\<mu\>><protectl|(>n<protectr|)>
  <protectlr|=> \<cal-S\><rprime|'>>.

  <\proof>
    \ \ intros n S <math|\<cal-E\><rsub|\<mu\>>> D2.

    \ \ induction D2.

    \ \ - intros <math|\<cal-E\><rsub|\<mu\>><rprime|'>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ exists <math|\<cal-S\><rprime|'>>.

    \ \ \ \ {\ 

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ - apply Fetch_1.

    \ \ \ \ }

    \ \ - intros <math|\<cal-E\><rsub|\<mu\>><rprime|'>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ apply IHD2 in H3.

    \ \ \ \ destruct H3.

    \ \ \ \ destruct H3.

    \ \ \ \ exists x.

    \ \ \ \ {

    \ \ \ \ \ \ split. \ \ \ 

    \ \ \ \ \ \ - exact H3.

    \ \ \ \ \ \ - apply Fetch_2.

    \ \ \ \ \ \ \ \ exact H4.

    \ \ \ \ }
  </proof>

  <with|font-series|bold|Theorem> <with|font-shape|italic|soundness> :
  \ <math|\<forall\>> <math|\<sigma\><rsub|1>>
  <math|<with|math-font|cal|\<sigma\><rsub|2>>>
  <math|\<sigma\><rsub|1><rprime|'>>, <math|\<sigma\><rsub|1>
  \<rightsquigarrow\><rsup|i> \<sigma\><rsub|2><space|1spc>\<rightarrow\><space|1spc>\<sigma\><rsub|1><rsup|\<star\>>=<rsub|\<sigma\>>
  \<sigma\><rsub|1><rprime|'><space|1spc>\<rightarrow\><space|1spc>\<exists\>\<sigma\><rsub|2><rprime|'>,<space|1spc>\<sigma\><rsub|1><rprime|'>
  \<rightsquigarrow\> \<sigma\><rsub|2><rprime|'><space|1spc>\<wedge\><space|1spc><with|math-font|cal|\<sigma\><rsub|2>><rsup|\<star\>>=<rsub|\<sigma\>>
  <with|math-font|cal|\<sigma\><rsub|2><rprime|'>>>.

  <\proof>
    \ \ intros <math|\<sigma\>>_1 <math|\<sigma\>>_2 <math|\<sigma\>>_1' D1.

    \ \ induction D1.

    \ \ - intro D2.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H9.

    \ \ \ \ inversion H14.

    \ \ \ \ assert (H222 := (fetch_sound _ _ H21)).

    \ \ \ \ subst.

    \ \ \ \ apply H222 in H0.

    \ \ \ \ destruct H0.

    \ \ \ \ destruct x.

    \ \ \ \ exists (<math|\<langle\>>t0, <math|\<cal-E\>>0,
    <math|\<cal-E\>>_<math|\<mu\>>0, <math|\<cal-S\><rprime|'>><math|\<rangle\>>).

    \ \ \ \ {\ 

    \ \ \ \ \ \ split.

    \ \ \ \ - apply K_var.

    \ \ \ \ \ \ inversion H24.

    \ \ \ \ \ \ inversion H.

    \ \ \ \ \ \ subst.

    \ \ \ \ \ \ assert (H71 := fetch_unique _ _ _ _ _ H7 H1).

    \ \ \ \ \ \ subst.

    \ \ \ \ \ \ assert (H82 := minus_unique _ _ _ _ H8 H2).

    \ \ \ \ \ \ subst.

    \ \ \ \ \ \ destruct H0.

    \ \ \ \ \ \ assumption.

    \ \ \ \ - subst.

    \ \ \ \ \ \ apply T_st.

    \ \ \ \ \ \ destruct H0.

    \ \ \ \ \ \ assumption. \ \ \ \ \ \ 

    \ \ \ \ \ \ assumption.

    \ \ \ \ }

    - \ intro D2.

    \ \ \ inversion D2.

    \ \ \ inversion H7.

    \ \ \ inversion H12.

    \ \ \ subst.

    \ \ \ exists (<math|\<langle\>>t', <math|\<cal-E\>>',
    <math|\<cal-E\>>_<math|\<mu\>>', [u', <math|\<cal-E\>>',
    <math|\<cal-E\>>_<math|\<mu\>>']::<math|\<cal-S\><rprime|'>><math|\<rangle\>>).

    \ \ \ {

    \ \ \ \ \ split.

    \ \ \ - apply K_app.

    \ \ \ - {\ 

    \ \ \ \ \ \ \ \ apply T_st.

    \ \ \ \ \ \ - apply T_cl.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ - apply T_stack_2.

    \ \ \ \ \ \ \ \ apply T_cl.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ }

    \ \ \ }

    - \ intro D2.

    \ \ \ inversion D2.

    \ \ \ inversion H7.

    \ \ \ inversion H12.

    \ \ \ inversion H8.

    \ \ \ subst.

    \ \ \ exists (<math|\<langle\>>t', (c'::<math|\<cal-E\>>'),
    <math|\<cal-E\>>_<math|\<mu\>>', <math|\<cal-S\><rprime|'>>0<math|\<rangle\>>).

    \ \ \ split.

    \ \ \ * apply K_abs.

    \ \ \ * apply T_st.

    \ \ \ \ \ \ + apply T_cl.

    \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ apply T_c_env_2.

    \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ + assumption.

    - \ intro D2.

    \ \ \ inversion D2.

    \ \ \ inversion H7.

    \ \ \ inversion H12.

    \ \ \ subst.

    \ \ \ exists (<math|\<langle\>>t', <math|\<cal-E\>>',
    (<math|\<cal-S\><rprime|'>>::<math|\<cal-E\>>_<math|\<mu\>>'),
    <math|\<cal-S\><rprime|'>><math|\<rangle\>>).

    \ \ \ split.

    \ \ \ * apply K_catch.

    \ \ \ * \ apply T_st.

    \ \ \ \ \ \ + apply T_cl.

    \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ apply T_k_env_2.\ 

    \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ + assumption.

    - \ intro D2.

    \ \ \ inversion D2.

    \ \ \ inversion H9.

    \ \ \ inversion H14.

    \ \ \ subst.

    \ \ \ assert (H333 := (fetch_mu_sound _ _ _ H0 _ H22)).

    \ \ \ destruct H333 as [ <math|\<cal-S\>2> H444 ].

    \ \ \ destruct H444.

    \ \ \ exists (<math|\<langle\>>t', <math|\<cal-E\>>',
    <math|\<cal-E\>>_<math|\<mu\>>', <math|\<cal-S\>2><math|\<rangle\>>).

    \ \ \ split.

    \ \ \ * apply K_throw.

    \ \ \ \ \ \ assumption.

    \ \ \ * apply T_st.

    \ \ \ \ \ \ + apply T_cl.

    \ \ \ \ \ \ \ \ \ assert (V2 := fetch_unique _ _ _ _ _ H25 H).

    \ \ \ \ \ \ \ \ \ subst.

    \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ + assumption.
  </proof>

  <\comment>
    <subsubsection|Completeness>
  </comment>

  <with|font-series|bold|Lemma> <with|font-shape|italic|fetch_complete> :
  \ <math|\<forall\>> <math|\<cal-E\>> <math|\<cal-E\><rprime|'>>,
  <math|\<cal-E\><rsup|\<star\>>=<rsub|e>
  \<cal-E\><rprime|'>\<rightarrow\>\<forall\><space|1spc>n<space|1spc>c<rprime|'>,<space|1spc>\<cal-E\><rprime|'><protectl|(>n<protectr|)>
  <protectlr|=> c<rprime|'>\<rightarrow\>\<exists\>c,c<rsup|\<star\>>=<rsub|c>
  c<rprime|'>\<wedge\>\<cal-E\><protectl|(>n<protectr|)> <protectlr|=> c>.

  <\proof>
    \ \ intros <math|\<cal-E\>> <math|\<cal-E\><rprime|'>> D1 n c D2.

    \ \ revert <math|\<cal-E\>> D1.

    \ \ induction D2.

    \ \ - intros <math|\<cal-E\>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ exists c.

    \ \ \ \ split.

    \ \ \ \ * assumption.

    \ \ \ \ * apply Fetch_1.

    \ \ - intros <math|\<cal-E\>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ apply IHD2 in H3.

    \ \ \ \ destruct H3.

    \ \ \ \ destruct H3.

    \ \ \ \ exists x.

    \ \ \ \ split. \ \ \ 

    \ \ \ \ * exact H3.

    \ \ \ \ * apply Fetch_2.

    \ \ \ \ \ \ exact H4.
  </proof>

  <with|font-series|bold|Lemma> <with|font-shape|italic|fetch_mu_complete> :
  \ <math|\<forall\>> <math|n> <math|\<cal-S\><rprime|'>>
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>>,
  <math|\<cal-E\><rsub|\<mu\>><rprime|'><protectl|(>n<protectr|)>
  <protectlr|=> \<cal-S\><rprime|'>\<rightarrow\>\<forall\>\<cal-E\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>><rsup|\<star\>>=<rsub|k>
  \<cal-E\><rsub|\<mu\>><rprime|'>\<rightarrow\>\<exists\>\<cal-S\>,\<cal-S\><rsup|\<star\>>=<rsub|s>
  \<cal-S\><rprime|'>\<wedge\>\<cal-E\><rsub|\<mu\>><protectl|(>n<protectr|)>
  <protectlr|=> \<cal-S\>>.

  <\proof>
    \ \ intros n S <math|\<cal-E\><rsub|\<mu\>><rprime|'>> D2.

    \ \ induction D2.

    \ \ - intros <math|\<cal-E\><rsub|\<mu\>>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ exists <math|\<cal-S\>>.

    \ \ \ \ {\ 

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ - apply Fetch_1.

    \ \ \ \ }

    \ \ - intros <math|\<cal-E\><rsub|\<mu\>>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ apply IHD2 in H3.

    \ \ \ \ destruct H3.

    \ \ \ \ destruct H3.

    \ \ \ \ exists x.

    \ \ \ \ {

    \ \ \ \ \ \ split. \ \ \ 

    \ \ \ \ \ \ - exact H3.

    \ \ \ \ \ \ - apply Fetch_2.

    \ \ \ \ \ \ \ \ exact H4.

    \ \ \ \ }
  </proof>

  <with|font-series|bold|Theorem> <with|font-shape|italic|completeness> :
  \ <math|\<forall\>> <math|\<sigma\><rsub|1><rprime|'>>
  <math|<with|math-font|cal|\<sigma\><rsub|2>><rprime|'>>
  <math|\<sigma\><rsub|1>>, <math|\<sigma\><rsub|1><rprime|'>
  \<rightsquigarrow\> \<sigma\><rsub|2><rprime|'><space|1spc>\<rightarrow\><space|1spc>\<sigma\><rsub|1><rsup|\<star\>>=<rsub|\<sigma\>>
  \<sigma\><rsub|1><rprime|'><space|1spc>\<rightarrow\><space|1spc>\<exists\>\<sigma\><rsub|2>,<space|1spc>\<sigma\><rsub|1>
  \<rightsquigarrow\><rsup|i> \<sigma\><rsub|2><space|1spc>\<wedge\><space|1spc><with|math-font|cal|\<sigma\><rsub|2>><rsup|\<star\>>=<rsub|\<sigma\>>
  <with|math-font|cal|\<sigma\><rsub|2><rprime|'>>>.

  <\proof>
    \ \ intros <math|\<sigma\>>_1' <math|\<sigma\>>_2' <math|\<sigma\>>_1 D1.

    \ \ induction D1.

    \ \ - intro D2.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H3.

    \ \ \ \ inversion H10.

    \ \ \ \ assert (H222 := (fetch_complete _ _ H17)).

    \ \ \ \ subst.

    \ \ \ \ apply H222 in H.

    \ \ \ \ destruct H.

    \ \ \ \ destruct x.

    \ \ \ \ exists (<math|\<langle\>>t0, n0, <math|\<cal-I\>>0,
    <math|\<cal-I\>>_<math|\<mu\>>0, <math|\<cal-E\>>1,
    <math|\<cal-E\>>_<math|\<mu\>>1, <math|\<cal-S\>>0<math|\<rangle\>>).

    \ \ \ \ {\ 

    \ \ \ \ \ \ split.

    \ \ \ \ - eapply I_var.

    \ \ \ \ \ \ exact H24.

    \ \ \ \ \ \ inversion H.

    \ \ \ \ \ \ subst.

    \ \ \ \ \ \ exact H1.

    \ \ \ \ - subst.

    \ \ \ \ \ \ apply T_st.

    \ \ \ \ \ \ destruct H.

    \ \ \ \ \ \ assumption. \ \ \ \ \ \ 

    \ \ \ \ \ \ assumption.

    \ \ \ \ }

    \ \ - intro D2.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H2.

    \ \ \ \ inversion H9.

    \ \ \ \ subst.

    \ \ \ \ exists (<math|\<langle\>>t2, n, <math|\<cal-I\>>,
    <math|\<cal-I\>>_<math|\<mu\>>, <math|\<cal-E\>>0,
    <math|\<cal-E\>>_<math|\<mu\>>0, [u2, n, <math|\<cal-I\>>,
    <math|\<cal-I\>>_<math|\<mu\>>, <math|\<cal-E\>>0,
    <math|\<cal-E\>>_<math|\<mu\>>0]::<math|\<cal-S\>>0<math|\<rangle\>>).

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - apply I_app.

    \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ apply T_st.

    \ \ \ \ \ \ \ \ \ \ - apply T_cl.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ - apply T_stack_2.

    \ \ \ \ \ \ \ \ \ \ \ \ apply T_cl.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }

    \ \ - intro D2.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H2.

    \ \ \ \ inversion H9.

    \ \ \ \ inversion H5.

    \ \ \ \ subst.

    \ \ \ \ exists (<math|\<langle\>>t2, (<math|S> n), (<math|S>
    n)::<math|\<cal-I\>>, <math|\<cal-I\>>_<math|\<mu\>>,
    (c0::<math|\<cal-E\>>0), <math|\<cal-E\>>_<math|\<mu\>>0,
    <math|\<cal-S\>>1<math|\<rangle\>>).

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - apply I_abs.

    \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ apply T_st.

    \ \ \ \ \ \ \ \ \ \ - apply T_cl.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ apply T_c_env_2.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }\ 

    \ \ - intro D2.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H2.

    \ \ \ \ inversion H9.

    \ \ \ \ subst.

    \ \ \ \ exists (<math|\<langle\>>t2, n, <math|\<cal-I\>>,
    (<math|\<cal-I\>>::<math|\<cal-I\>>_<math|\<mu\>>), <math|\<cal-E\>>0,
    (<math|\<cal-S\>>0::<math|\<cal-E\>>_<math|\<mu\>>0),
    <math|\<cal-S\>0><math|\<rangle\>>).

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - apply I_catch.

    \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ apply T_st.

    \ \ \ \ \ \ \ \ \ \ - apply T_cl.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ apply T_k_env_2.\ 

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }

    \ \ - intro D2.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H3.

    \ \ \ \ inversion H10.

    \ \ \ \ subst.

    \ \ \ \ assert (H333 := (fetch_mu_complete _ _ _ H _ H18)).

    \ \ \ \ destruct H333 as [ <math|\<cal-S\>2> H444 ].

    \ \ \ \ destruct H444.

    \ \ \ \ exists (<math|\<langle\>>t2, n, <math|\<cal-I\>>',
    <math|\<cal-I\>>_<math|\<mu\>>, <math|\<cal-E\>>0,
    <math|\<cal-E\>>_<math|\<mu\>>0, <math|\<cal-S\>2><math|\<rangle\>>).

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - apply I_throw.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ apply T_st.

    \ \ \ \ \ \ \ \ \ \ - apply T_cl.

    \ \ \ \ \ \ \ \ \ \ \ \ subst.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }
  </proof>

  <\comment>
    <subsection|Lock-step simulation <math|<around*|(|-|)><rsup|\<diamond\>>>>
  </comment>

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<cal-E\>> <protectlr|(> <math|n> <protect_l|<math|->> \ <math|k>
  <protectlr|)> <protectlr|=> <math|c>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|compute_l> :
  <with|font-shape|italic|c_env_i> <math|\<rightarrow\>>
  <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|clos_i> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \;

  \| <with|font-shape|italic|Compute_l<math|<rsub|1>>> : <math|\<forall\>>
  <math|\<cal-E\>> <math|n> <math|g> <math|k> <math|c>,

  <\equation*>
    <dfrac|n<protectlr|<math|->>k <protectlr|=>
    g<value|sep>\<cal-E\><protectl|(>g<protectr|)><protectlr|=>
    c|\<cal-E\><protectlr|(><math|n> <protectl|<math|->> k<protectlr|)>
    <protectlr|=> c>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<cal-E\>> <protectlr|(> <math|n>
  <protect_l|<math|->> \ <math|k> <protectlr|)> <protectlr|=> <math|c>" :=
  (<with|font-shape|italic|compute_l> <math|\<cal-E\>> <math|n> <math|k>
  <math|c>).

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|compute_l_unique> :
  <math|\<forall\>> <math|\<cal-E\>> <math|n> <math|k> <math|c>
  <math|c<rprime|'>>, <math|\<cal-E\><protectlr|(><math|n>
  <protectl|<math|->> k<protectlr|)> <protectlr|=> c> <math|\<rightarrow\>>
  <math|\<cal-E\><protectlr|(><math|n> <protectl|<math|->> k<protectlr|)>
  <protectlr|=> c<rprime|'>> <math|\<rightarrow\>> <math|c=c<rprime|'>>.

  <\proof>
    \ \ introv. revert_last.

    \ \ intros y1 H H0. revert H0. revert y1.

    \ \ induction H;\ 

    \ \ \ \ (intros y2 D2; inversion D2; repeat progress ( eauto \|\|
    econstructor \|\| gen_subst \|\| crush )).
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|compute_l_unique> :
  <with|font-shape|italic|gen_subst_db>.

  \;

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|<with|math-font|cal|c>> '^<math|\<diamond\>>=_<math|c>'
  <math|<with|math-font|cal|c><rprime|'>>" (at level 30).

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<cal-S\>> '^<math|\<diamond\>>=_<math|s>'
  <math|\<cal-S\><rprime|'>>" (at level 30).

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<cal-E\><rsub|\<mu\>>> '^<math|\<diamond\>>=_<math|k>'
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|flatten> :
  <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|c_env_i> <math|\<rightarrow\>>
  <with|font-shape|italic|vector> <math|\<rightarrow\>>
  <with|font-shape|italic|l_env_l> <math|\<rightarrow\>><with|font-shape|italic|Prop>
  :=

  \| <with|font-shape|italic|T_flatten<math|<rsub|1>>> : <math|\<forall\>>
  <math|n> <math|\<cal-E\>>,

  <\equation*>
    <math-it|flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc><math-it|nil><space|1spc>
    <math-it|nil>
  </equation*>

  \;

  \| <with|font-shape|italic|T_flatten<math|<rsub|2>>> : <math|\<forall\>>
  <math|c> <math|c<rprime|'>> <math|n> <math|k> <math|\<cal-E\>>
  <math|\<cal-I\>> <math|\<cal-L\>>,

  <\equation*>
    <dfrac|\<cal-E\><protectlr|(><math|n> <protectl|<math|->> k<protectlr|)>
    <protectlr|=> c<value|sep>c<rsup|\<diamond\>>=<rsub|c>
    <with|math-font|cal|c<rprime|'>><value|sep><math-it|flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><space|1spc>\<cal-L\>|<math-it|flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc><around*|(|k
    ::\<cal-I\>|)><space|1spc> <around*|(|c<rprime|'> ::\<cal-L\>|)>>
  </equation*>

  \;

  <with|font-series|bold|with> <with|font-shape|italic|map_flatten> :
  <with|font-shape|italic|nat> <math|\<rightarrow\>>
  <with|font-shape|italic|c_env_i> <math|\<rightarrow\>>
  <with|font-shape|italic|table> <math|\<rightarrow\>>
  <with|font-shape|italic|m_env_l> <math|\<rightarrow\>><with|font-shape|italic|Prop>
  :=

  \| <with|font-shape|italic|T_map_flatten<math|<rsub|1>>> :
  <math|\<forall\>> <math|n> <math|\<cal-E\>>,

  <\equation*>
    <math-it|map_flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc><math-it|nil>
    <math-it|nil>
  </equation*>

  \;

  \| <with|font-shape|italic|T_map_flatten<math|<rsub|2>>> :
  <math|\<forall\>> <math|n> <math|\<cal-E\>> <math|\<cal-I\>>
  <math|\<cal-L\>> <math|\<cal-I\><rsub|\<mu\>>>
  <math|\<cal-L\><rsub|\<mu\>>>,

  <\equation*>
    <dfrac|<math-it|flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><space|1spc>\<cal-L\><value|sep><math-it|map_flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><rsub|\<mu\>><space|1spc>\<cal-L\><rsub|\<mu\>>|<math-it|map_flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc><around*|(|\<cal-I\>
    ::\<cal-I\><rsub|\<mu\>>|)><space|1spc><around*|(|\<cal-L\>
    ::\<cal-L\><rsub|\<mu\>>|)>>
  </equation*>

  \;

  <with|font-series|bold|with> <with|font-shape|italic|translate_clos_l> :
  <with|font-shape|italic|clos_i> <math|\<rightarrow\>>
  <with|font-shape|italic|clos_l> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \| <with|font-shape|italic|T_cl_l> : <math|\<forall\>> <math|n> <math|t>
  <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|\<cal-L\>>
  <math|\<cal-L\><rsub|\<mu\>>> <math|\<cal-E\>>
  <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-E\><rsub|\<mu\>><rprime|'>>,

  <\equation*>
    <dfrac|<math-it|flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><space|1spc>\<cal-L\><value|sep><math-it|map_flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><rsub|\<mu\>><space|1spc>\<cal-L\><rsub|\<mu\>><value|sep>\<cal-E\><rsub|\<mu\>><rsup|\<diamond\>>=<rsub|k>
    \<cal-E\><rsub|\<mu\>><rprime|'>|<around*|[|t,n,
    \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>|]><rsup|\<diamond\>>=<rsub|c>
    <around*|[| t,\<cal-L\>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>><rprime|'>|]>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|<with|math-font|cal|c>>
  '^<math|\<diamond\>>=_<math|c>' <math|c<rprime|'>>" :=
  (<with|font-shape|italic|translate_clos_l> <math|<with|math-font|cal|c>>
  <math|c<rprime|'>>)

  \;

  <with|font-series|bold|with> <with|font-shape|italic|translate_stack_l> :
  <with|font-shape|italic|stack_i> <math|\<rightarrow\>>
  <with|font-shape|italic|stack_l> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \| <with|font-shape|italic|T_stack_l<math|<rsub|1>>> :

  <\equation*>
    <math-it|nil> <rsup|\<diamond\>>=<rsub|s> <math-it|nil>
  </equation*>

  \;

  \| <with|font-shape|italic|T_stack_l<math|<rsub|2>>> : <math|\<forall\>>
  <math|c> <math|c<rprime|'>> <math|\<cal-S\>> <math|\<cal-S\><rprime|'>>,

  <\equation*>
    <dfrac|c<rsup|\<diamond\>>=<rsub|c> c<rprime|'><value|sep>\<cal-S\><rsup|\<diamond\>>=<rsub|s>
    \<cal-S\><rprime|'>|<with|math-font|cal|<around*|(|c
    ::\<cal-S\>|)>><rsup|\<diamond\>>=<rsub|s>
    <around*|(|<with|math-font|cal|c<rprime|'> ::\<cal-S\><rprime|'>>|)>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<cal-S\>>
  '^<math|\<diamond\>>=_<math|s>' <math|\<cal-S\><rprime|'>>" :=
  (<with|font-shape|italic|translate_stack_l> <math|\<cal-S\>>
  <math|\<cal-S\><rprime|'>>)

  \;

  <with|font-series|bold|with> <with|font-shape|italic|translate_k_env_l> :
  <with|font-shape|italic|k_env_i> <math|\<rightarrow\>>
  <with|font-shape|italic|k_env_l> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \| <with|font-shape|italic|T_k_env_l<math|<rsub|1>>> :

  <\equation*>
    <math-it|nil> <rsup|\<diamond\>>=<rsub|k> <math-it|nil>
  </equation*>

  \;

  \| <with|font-shape|italic|T_k_env_l<math|<rsub|2>>> : <math|\<forall\>>
  <math|\<cal-S\>> <math|\<cal-S\><rprime|'>> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>>,

  <\equation*>
    <dfrac|\<cal-S\><rsup|\<diamond\>>=<rsub|s>
    \<cal-S\><rprime|'><value|sep>\<cal-E\><rsub|\<mu\>><rsup|\<diamond\>>=<rsub|k>
    \<cal-E\><rsub|\<mu\>><rprime|'>|<with|math-font|cal|<around*|(|\<cal-S\>
    ::\<cal-E\><rsub|\<mu\>>|)>><rsup|\<diamond\>>=<rsub|k>
    <around*|(|\<cal-S\><rprime|'> ::\<cal-E\><rsub|\<mu\>><rprime|'>|)>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<cal-E\><rsub|\<mu\>>>
  '^<math|\<diamond\>>=_<math|k>' <math|\<cal-E\><rsub|\<mu\>><rprime|'>>" :=
  (<with|font-shape|italic|translate_k_env_l> <math|\<cal-E\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>>).

  \;

  \;

  <with|font-series|bold|Reserved> <with|font-series|bold|Notation>
  "<math|\<sigma\>> '^<math|\<diamond\>>=_<math|\<sigma\>>'
  <math|\<sigma\><rprime|'>>" (at level 30).

  \;

  <with|font-series|bold|Inductive> <with|font-shape|italic|translate_st_l> :
  <with|font-shape|italic|state_i> <math|\<rightarrow\>>
  <with|font-shape|italic|state_l> <math|\<rightarrow\>>
  <with|font-shape|italic|Prop> :=

  \| <with|font-shape|italic|T_st_l> : <math|\<forall\>> <math|n> <math|t>
  <math|u> <math|\<cal-I\>> <math|\<cal-I\><rsub|\<mu\>>> <math|\<cal-E\>>
  <math|\<cal-L\>> <math|\<cal-L\><rsub|\<mu\>>>
  <math|\<cal-E\><rsub|\<mu\>>> <math|\<cal-E\><rsub|\<mu\>><rprime|'>>
  <math|\<cal-S\>> <math|\<cal-S\><rprime|'>>,

  <\equation*>
    <dfrac|<around*|[| t,n, \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>|]><rsup|\<diamond\>>=<rsub|c>
    <around*|[|u,\<cal-L\>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>><rprime|'>|]><value|sep>\<cal-S\><rsup|\<diamond\>>=<rsub|s>
    \<cal-S\><rprime|'>|\<langle\> t,n, \<cal-I\>,\<cal-I\><rsub|\<mu\>>,\<cal-E\>,\<cal-E\><rsub|\<mu\>>,\<cal-S\>\<rangle\><rsup|\<diamond\>>=<rsub|\<sigma\>>
    \<langle\> u,\<cal-L\>,\<cal-L\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>><rprime|'>,\<cal-S\><rprime|'>\<rangle\>>
  </equation*>

  \;

  <with|font-series|bold|where> "<math|\<sigma\>>
  '^<math|\<diamond\>>=_<math|\<sigma\>>' <math|\<sigma\><rprime|'>>" :=
  (<with|font-shape|italic|translate_st_l> <math|\<sigma\>>
  <math|\<sigma\><rprime|'>>).

  \;

  <with|font-series|bold|Scheme> <with|font-shape|italic|flatten_mut> :=
  <with|font-series|bold|Induction> <with|font-series|bold|for>
  <with|font-shape|italic|flatten> <with|font-series|bold|Sort>
  <with|font-shape|italic|Prop>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|map_flatten_mut> :=
  <with|font-series|bold|Induction> <with|font-series|bold|for>
  <with|font-shape|italic|map_flatten> <with|font-series|bold|Sort>
  <with|font-shape|italic|Prop>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_clos_l_mut> :=
  <with|font-series|bold|Induction> <with|font-series|bold|for>
  <with|font-shape|italic|translate_clos_l> <with|font-series|bold|Sort>
  <with|font-shape|italic|Prop>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_stack_l_mut> :=
  <with|font-series|bold|Induction> <with|font-series|bold|for>
  <with|font-shape|italic|translate_stack_l> <with|font-series|bold|Sort>
  <with|font-shape|italic|Prop>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_k_env_l_mut> :=
  <with|font-series|bold|Induction> <with|font-series|bold|for>
  <with|font-shape|italic|translate_k_env_l> <with|font-series|bold|Sort>
  <with|font-shape|italic|Prop>.

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|flatten_unique> :
  <math|\<forall\>> <math|n> <math|\<cal-E\>> <math|\<cal-I\>>
  <math|\<cal-L\>> <math|\<cal-L\><rprime|'>>,
  <math|<math-it|flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><space|1spc>\<cal-L\>\<rightarrow\><math-it|flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><space|1spc>\<cal-L\><rprime|'>\<rightarrow\>\<cal-L\>=\<cal-L\><rprime|'>
  >

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|map_flatten_unique> : <math|\<forall\>> <math|n>
  <math|\<cal-E\>> <math|\<cal-I\><rsub|\<mu\>>>
  <math|\<cal-L\><rsub|\<mu\>>> <math|\<cal-L\><rsub|\<mu\>><rprime|'>>,\ 

  <space|6em><math|<math-it|map_flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><rsub|\<mu\>><space|1spc>\<cal-L\><rsub|\<mu\>>\<rightarrow\><math-it|map_flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><rsub|\<mu\>><space|1spc>\<cal-L\><rsub|\<mu\>><rprime|'>\<rightarrow\>\<cal-L\><rsub|\<mu\>>=\<cal-L\><rsub|\<mu\>><rprime|'>
  >

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_clos_l_unique> : <math|\<forall\>>
  <math|c> <math|c<rsub|2>> <math|c<rsub|2><rprime|'>>,
  <math|c<rsup|\<diamond\>>=<rsub|c> c<rsub|2>> <math|\<rightarrow\>>
  <math|c<rsup|\<diamond\>>=<rsub|c> c<rsub|2><rprime|'>>
  <math|\<rightarrow\>> <math|c<rsub|2>=c<rsub|2><rprime|'>>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_stack_l_unique> : <math|\<forall\>>
  <math|s> <math|s<rsub|2>> <math|s<rsub|2><rprime|'>>,
  <math|s<rsup|\<diamond\>>=<rsub|s> s<rsub|2>> <math|\<rightarrow\>>
  <math|s<rsup|\<diamond\>>=<rsub|s> s<rsub|2><rprime|'>>
  <math|\<rightarrow\>> <math|s<rsub|2>=s<rsub|2><rprime|'>>

  <space|1.5em><with|font-series|bold|with>
  <with|font-shape|italic|translate_k_env_l_unique> : <math|\<forall\>>
  <math|k> <math|k<rsub|2>> <math|k<rsub|2><rprime|'>>,
  <math|k<rsup|\<diamond\>>=<rsub|k> k<rsub|2>> <math|\<rightarrow\>>
  <math|k<rsup|\<diamond\>>=<rsub|k> k<rsub|2><rprime|'>>
  <math|\<rightarrow\>> <math|k<rsub|2>=k<rsub|2><rprime|'>>.

  <\proof>
    \ \ <with|font-series|bold|Tactic Notation>
    <verbatim|"><with|font-shape|italic|translate_l_unique_tac><verbatim|">
    tactic(L) :=

    \ \ \ \ repeat clear_last;

    \ \ \ \ introv; revert_last;

    \ \ \ \ intros y1 H H0; revert H0; revert y1;

    \ \ \ \ apply (L

    \ \ \ \ \ \ \ \ \ \ (<with|font-series|bold|fun> <math|n>
    <math|\<cal-E\>> <math|\<cal-I\>> <math|\<cal-L\>> <math|h>
    <verbatim|=\<gtr\>> <math|\<forall\>> <math|\<cal-L\><rprime|'>>,
    <math|<math-it|flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><space|1spc>\<cal-L\><rprime|'>\<rightarrow\>\<cal-L\>=\<cal-L\><rprime|'>>)

    \ \ \ \ \ \ \ \ \ \ (<with|font-series|bold|fun> <math|n>
    <math|\<cal-E\>> <math|\<cal-I\><rsub|\<mu\>>>
    <math|\<cal-L\><rsub|\<mu\>>> <math|h> <verbatim|=\<gtr\>>
    <math|\<forall\>> <math|\<cal-L\><rsub|\<mu\>><rprime|'>>,
    <math|<math-it|map_flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><rsub|\<mu\>><space|1spc>\<cal-L\><rsub|\<mu\>><rprime|'>\<rightarrow\>\<cal-L\><rsub|\<mu\>>=\<cal-L\><rsub|\<mu\>><rprime|'>>)

    \ \ \ \ \ \ \ \ \ \ (<with|font-series|bold|fun> <math|c>
    <math|c<rsub|2>> <math|h> <verbatim|=\<gtr\>> <math|\<forall\>>
    <math|c<rsub|2><rprime|'>>, <math|c<rsup|\<diamond\>>=<rsub|c>
    c<rsub|2><rprime|'>> <math|\<rightarrow\>>
    <math|c<rsub|2>=c<rsub|2><rprime|'>>)

    \ \ \ \ \ \ \ \ \ \ (<with|font-series|bold|fun> <math|s>
    <math|s<rsub|2>> <math|h> <verbatim|=\<gtr\>> <math|\<forall\>>
    <math|s<rsub|2><rprime|'>>, <math|s<rsup|\<diamond\>>=<rsub|s>
    s<rsub|2><rprime|'>> <math|\<rightarrow\>>
    <math|s<rsub|2>=s<rsub|2><rprime|'>>)

    \ \ \ \ \ \ \ \ \ \ (<with|font-series|bold|fun> <math|k>
    <math|k<rsub|2>> <math|h> <verbatim|=\<gtr\>> <math|\<forall\>>
    <math|k<rsub|2><rprime|'>>, <math|k<rsup|\<diamond\>>=<rsub|k>
    k<rsub|2><rprime|'>> <math|\<rightarrow\>>
    <math|k<rsub|2>=k<rsub|2><rprime|'>>));

    \ \ \ \ (intros; revert_last; intro D2; inversion D2;

    \ \ \ \ \ repeat progress ( eauto \|\| econstructor \|\| gen_subst \|\|
    crush )).

    \ \ - <with|font-shape|italic|translate_l_unique_tac>
    <with|font-shape|italic|flatten_mut>.

    \ \ - <with|font-shape|italic|translate_l_unique_tac>
    <with|font-shape|italic|map_flatten_mut>.

    \ \ - <with|font-shape|italic|translate_l_unique_tac>
    <with|font-shape|italic|translate_clos_l_mut>.

    \ \ - <with|font-shape|italic|translate_l_unique_tac>
    <with|font-shape|italic|translate_stack_l_mut>.

    \ \ - <with|font-shape|italic|translate_l_unique_tac>
    <with|font-shape|italic|translate_k_env_l_mut>.
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|flatten_unique> :
  <with|font-shape|italic|gen_subst_db>.

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|map_flatten_unique> :
  <with|font-shape|italic|gen_subst_db>.

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|translate_clos_l_unique> :
  <with|font-shape|italic|gen_subst_db>.

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|translate_stack_l_unique> :
  <with|font-shape|italic|gen_subst_db>.

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|translate_k_env_l_unique> :
  <with|font-shape|italic|gen_subst_db>.

  \;

  <with|font-series|bold|Lemma> <with|font-shape|italic|translate_st_l_unique>
  : <math|\<forall\>> <math|\<sigma\>> <math|\<sigma\><rsub|2>>
  <math|\<sigma\><rsub|2><rprime|'>>, <math|\<sigma\><rsup|\<diamond\>>=<rsub|\<sigma\>>
  \<sigma\><rsub|2>> <math|\<rightarrow\>>
  <math|\<sigma\><rsup|\<diamond\>>=<rsub|\<sigma\>>
  \<sigma\><rsub|2><rprime|'>> <math|\<rightarrow\>>
  <math|\<sigma\><rsub|2>=\<sigma\><rsub|2><rprime|'>>.

  <\proof>
    \ \ introv. revert_last.

    \ \ intros y1 H H0. revert H0. revert y1.

    \ \ induction H;\ 

    \ \ \ \ (intros y2 D2; inversion D2; repeat progress ( eauto \|\|
    econstructor \|\| gen_subst \|\| crush )).
  </proof>

  <with|font-series|bold|Hint> <with|font-series|bold|Resolve>
  <with|font-shape|italic|translate_st_l_unique> :
  <with|font-shape|italic|gen_subst_db>.

  <\comment>
    <subsubsection|Soundness>
  </comment>

  <with|font-series|bold|Lemma> <with|font-shape|italic|minus_next> :
  <math|\<forall\>> <math|n> <math|k> <math|g>, <math|n<protectlr|<math|->>k
  <protectlr|=> g> <math|\<rightarrow\>> <math|S n<protectlr|<math|->>k
  <protectlr|=> S g>.

  <\proof>
    \ \ intros <math|n> <math|k> <math|g> <math|\<cal-D\><rsub|1>>.

    \ \ induction <math|\<cal-D\><rsub|1>>.

    \ \ - apply Minus_1.

    \ \ - apply Minus_2.

    \ \ \ \ assumption.
  </proof>

  <with|font-series|bold|Lemma> <with|font-shape|italic|weaken_flatten> :
  <math|\<forall\>> {<math|n> <math|\<cal-E\>> <math|\<cal-I\>>
  <math|\<cal-L\>>}, <math|\<forall\>> <math|c<rprime|'>>,
  <with|font-shape|italic|flatten> <math|n> <math|\<cal-E\>> <math|\<cal-I\>>
  <math|\<cal-L\>> <math|\<rightarrow\>> <with|font-shape|italic|flatten>
  (<math|S n>) <math|<around*|(|c<rprime|'> ::\<cal-E\>|)>> <math|\<cal-I\>>
  <math|\<cal-L\>>.

  <\proof>
    \ \ intros <math|n> <math|\<cal-E\>> <math|\<cal-I\>> <math|\<cal-L\>>
    <math|c<rprime|'>> <math|\<cal-D\>>_1.

    \ \ induction <math|\<cal-D\>>_1.

    \ \ - apply T_flatten_1.

    \ \ - inversion_clear H.

    \ \ \ \ assert (<math|\<cal-D\>>_4 := minus_next _ _ _ H1).

    \ \ \ \ {

    \ \ \ \ \ \ eapply T_flatten_2.

    \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ eapply Compute_l_1.

    \ \ \ \ \ \ \ \ \ \ - exact <math|\<cal-D\>>_4.

    \ \ \ \ \ \ \ \ \ \ - apply Fetch_2.\ 

    \ \ \ \ \ \ \ \ \ \ \ \ exact H2.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ - assumption.

    \ \ \ \ }
  </proof>

  <with|font-series|bold|Lemma> <with|font-shape|italic|weaken_map_flatten> :
  <math|\<forall\>> {<math|n> <math|\<cal-E\>> <math|\<cal-I\><rsub|\<mu\>>>
  <math|\<cal-L\><rsub|\<mu\>>>}, <math|\<forall\>> <math|c<rprime|'>>,

  <space|2em><with|font-shape|italic|map_flatten> <math|n> <math|\<cal-E\>>
  <math|\<cal-I\><rsub|\<mu\>>> <math|\<cal-L\><rsub|\<mu\>>>
  <math|\<rightarrow\>> <with|font-shape|italic|map_flatten> (<math|S n>)
  <math|<around*|(|c<rprime|'> ::\<cal-E\>|)>> <math|\<cal-I\><rsub|\<mu\>>>
  <math|\<cal-L\><rsub|\<mu\>>>.

  <\proof>
    \ \ intros <math|n> <math|\<cal-E\>> <math|\<cal-I\><rsub|\<mu\>>>
    <math|\<cal-L\><rsub|\<mu\>>> <math|c<rprime|'>> <math|\<cal-D\>>_1.

    \ \ induction <math|\<cal-D\>>_1.

    \ \ - apply T_map_flatten_1.

    \ \ - eapply weaken_flatten in H.

    \ \ \ \ {

    \ \ \ \ \ \ apply \ T_map_flatten_2.

    \ \ \ \ \ \ - exact H.

    \ \ \ \ \ \ - assumption.

    \ \ \ \ }
  </proof>

  <with|font-series|bold|Lemma> <with|font-shape|italic|map_flatten_sound> :
  <math|\<forall\>> <math|n> <math|\<cal-E\>> <math|\<cal-I\><rprime|'>>
  <math|\<cal-I\><rsub|\<mu\>>> <math|\<alpha\>>,\ 

  <space|2em><math|\<cal-I\><rsub|\<mu\>><protectl|(>\<alpha\><protectr|)><protectlr|=>
  \<cal-I\><rprime|'>> <math|\<rightarrow\>> <math|\<forall\>>
  <math|\<cal-L\><rsub|\<mu\>>>, <with|font-shape|italic|map_flatten>
  <math|n> <math|\<cal-E\>> <math|\<cal-I\><rsub|\<mu\>>>
  <math|\<cal-L\><rsub|\<mu\>>> <math|\<rightarrow\>> <math|\<exists\>>
  <math|\<cal-L\><rprime|'>>, <with|font-shape|italic|flatten> <math|n>
  <math|\<cal-E\>> <math|\<cal-I\><rprime|'>> <math|\<cal-L\><rprime|'>>
  <math|\<wedge\>> \ <math|\<cal-L\><rsub|\<mu\>><protectl|(>\<alpha\><protectr|)><protectlr|=>
  \<cal-L\><rprime|'>>.

  <\proof>
    \ \ intros n <math|\<cal-E\>> <math|\<cal-I\>>'
    <math|\<cal-I\>>_<math|\<mu\>> <math|\<alpha\>> <math|\<cal-D\>>_2.

    \ \ induction <math|\<cal-D\>>_2.

    \ \ - intros <math|\<cal-L\>>_<math|\<mu\>> <math|\<cal-D\>>_1.

    \ \ \ \ inversion <math|\<cal-D\>>_1; subst.

    \ \ \ \ exists <math|\<cal-L\>>.

    \ \ \ \ split.

    \ \ \ \ exact H3.

    \ \ \ \ apply Fetch_1.

    \ \ - intros <math|\<cal-L\>>_<math|\<mu\>> <math|\<cal-D\>>_1.

    \ \ \ \ inversion <math|\<cal-D\>>_1; subst.

    \ \ \ \ assert (D34 := IH<math|\<cal-D\>>_2 _ H5).

    \ \ \ \ destruct D34.

    \ \ \ \ exists x.

    \ \ \ \ destruct H.

    \ \ \ \ split.

    \ \ \ \ exact H.

    \ \ \ \ apply Fetch_2.

    \ \ \ \ assumption.
  </proof>

  <with|font-series|bold|Lemma> <with|font-shape|italic|fetch_sound_l> :
  \ <math|\<forall\>> <math|n> <math|l> <math|k> <math|c> <math|\<cal-I\>>,\ 

  <space|2em><math|\<cal-I\><protectl|(>l<protectr|)> <protectlr|=>
  k\<rightarrow\>\<forall\>\<cal-E\>,\<cal-E\><protectlr|(><math|n>
  <protectl|<math|->> k<protectlr|)> <protectlr|=>
  c\<rightarrow\>\<forall\>\<cal-L\>,<math-it|flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><space|1spc>\<cal-L\>\<rightarrow\>\<exists\>c<rprime|'>,c<rsup|\<diamond\>>=<rsub|c>
  c<rprime|'>\<wedge\>\<cal-L\><protectl|(>l<protectr|)> <protectlr|=>
  c<rprime|'>>.

  <\proof>
    \ \ intros <math|n> <math|l> <math|k> <math|c> <math|\<cal-I\>> D1.

    \ \ induction D1.

    \ \ - intros <math|\<cal-E\>>' D2 <math|\<cal-L\>> D3.

    \ \ \ \ inversion D3.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H1.

    \ \ \ \ subst.

    \ \ \ \ exists c'.

    \ \ \ \ assert (H666 := minus_unique _ _ _ _ H7 H13).

    \ \ \ \ subst.\ 

    \ \ \ \ assert (H555 := fetch_unique _ _ _ _ _ H8 H14).

    \ \ \ \ subst.\ 

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ - apply Fetch_1.

    \ \ \ \ }

    \ \ - intros <math|\<cal-E\>>' D2 <math|\<cal-L\>> D3.

    \ \ \ \ inversion D3.

    \ \ \ \ subst.

    \ \ \ \ assert (H888 := IHD1 _ D2 _ H6).

    \ \ \ \ destruct H888.

    \ \ \ \ destruct H.

    \ \ \ \ exists x.

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ - apply Fetch_2.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ }
  </proof>

  <with|font-series|bold|Lemma> <with|font-shape|italic|fetch_mu_sound_l> :
  \ <math|\<forall\>> <math|n> <math|\<cal-S\>>
  <math|\<cal-E\><rsub|\<mu\>>>, <math|\<cal-E\><rsub|\<mu\>><protectl|(>n<protectr|)>
  <protectlr|=> \<cal-S\>\<rightarrow\>\<forall\>\<cal-E\><rprime|'><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>><rsup|\<diamond\>>=<rsub|k>
  \<cal-E\><rprime|'><rsub|\<mu\>>\<rightarrow\>\<exists\>\<cal-S\><rprime|'>,\<cal-S\><rsup|\<diamond\>>=<rsub|s>
  \<cal-S\><rprime|'>\<wedge\>\<cal-E\><rprime|'><rsub|\<mu\>><protectl|(>n<protectr|)>
  <protectlr|=> \<cal-S\><rprime|'>>.

  <\proof>
    \ \ intros n S <math|\<cal-E\><rsub|\<mu\>>> D2.

    \ \ induction D2.

    \ \ - intros <math|\<cal-E\><rsub|\<mu\>><rprime|'>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ exists <math|\<cal-S\><rprime|'>>.

    \ \ \ \ {\ 

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ - apply Fetch_1.

    \ \ \ \ }

    \ \ - intros <math|\<cal-E\><rsub|\<mu\>><rprime|'>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ apply IHD2 in H3.

    \ \ \ \ destruct H3.

    \ \ \ \ destruct H3.

    \ \ \ \ exists x.

    \ \ \ \ {

    \ \ \ \ \ \ split. \ \ \ 

    \ \ \ \ \ \ - exact H3.

    \ \ \ \ \ \ - apply Fetch_2.

    \ \ \ \ \ \ \ \ exact H4.

    \ \ \ \ }
  </proof>

  <with|font-series|bold|Theorem> <with|font-shape|italic|soundness_l> :
  \ <math|\<forall\>> <math|\<sigma\><rsub|1>>
  <math|<with|math-font|cal|\<sigma\><rsub|2>>>
  <math|\<sigma\><rsub|1><rprime|'>>, <math|\<sigma\><rsub|1>
  \<rightsquigarrow\><rsup|i> \<sigma\><rsub|2><space|1spc>\<rightarrow\><space|1spc>\<sigma\><rsub|1><rsup|\<diamond\>>=<rsub|\<sigma\>>
  \<sigma\><rsub|1><rprime|'><space|1spc>\<rightarrow\><space|1spc>\<exists\>\<sigma\><rsub|2><rprime|'>,<space|1spc>\<sigma\><rsub|1><rprime|'>
  \<rightsquigarrow\><rsup|l> \<sigma\><rsub|2><rprime|'><space|1spc>\<wedge\><space|1spc><with|math-font|cal|\<sigma\><rsub|2>><rsup|\<diamond\>>=<rsub|\<sigma\>>
  <with|math-font|cal|\<sigma\><rsub|2><rprime|'>>>.

  <\proof>
    \ \ intros <math|\<sigma\>>_1 <math|\<sigma\>>_2 <math|\<sigma\>>_1' D1.

    \ \ induction D1.

    \ \ - intro D2.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H9.

    \ \ \ \ inversion H.

    \ \ \ \ subst.

    \ \ \ \ assert (H111 := Compute_l_1 _ _ _ _ _ H25 H0).

    \ \ \ \ assert (H222 := fetch_sound_l _ _ _ _ _ H24 _ H111 _ H14).

    \ \ \ \ destruct H222.

    \ \ \ \ destruct H1.

    \ \ \ \ destruct x.

    \ \ \ \ exists <math|(\<langle\>t0, \<cal-L\>0, \<cal-L\>_\<mu\>0,
    \<cal-E\>_\<mu\>0, \<cal-S\><rprime|'>\<rangle\>)>.

    \ \ \ \ {\ 

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - apply L_var.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ - subst.

    \ \ \ \ \ \ \ \ apply T_st_l.

    \ \ \ \ \ \ \ \ destruct H0.

    \ \ \ \ \ \ \ \ assumption. \ \ \ \ \ \ 

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ }

    \ \ - intro D2.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H7.

    \ \ \ \ subst.

    \ \ \ \ exists <math|(\<langle\>t, \<cal-L\>, \<cal-L\>_\<mu\>,
    \<cal-E\>_\<mu\><rprime|'>, [u, \<cal-L\>, \<cal-L\>_\<mu\>,
    \<cal-E\>_\<mu\><rprime|'>]::\<cal-S\><rprime|'>\<rangle\>)>.

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - apply L_app.

    \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ apply T_st_l.

    \ \ \ \ \ \ \ \ \ \ - apply T_cl_l.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ - apply T_stack_l_2.

    \ \ \ \ \ \ \ \ \ \ \ \ apply T_cl_l.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ }

    \ \ \ }

    \ \ - intro D2.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H7.

    \ \ \ \ inversion H8.

    \ \ \ \ subst.

    \ \ \ \ exists <math|(\<langle\>t, (c<rprime|'>::\<cal-L\>),
    \<cal-L\>_\<mu\>, \<cal-E\>_\<mu\><rprime|'>,
    \<cal-S\><rprime|'>0\<rangle\>)>.

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - apply L_abs.

    \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ apply T_st_l.

    \ \ \ \ \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_cl_l.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ eapply T_flatten_2.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ eapply Compute_l_1.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - apply minus_id.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - apply Fetch_1.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - apply weaken_flatten.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - apply weaken_map_flatten.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }

    \ \ - intro D2.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H7.

    \ \ \ \ subst.

    \ \ \ \ exists <math|(\<langle\>t, \<cal-L\>,
    (\<cal-L\>::\<cal-L\>_\<mu\>), (\<cal-S\><rprime|'>::\<cal-E\>_\<mu\><rprime|'>),
    \<cal-S\><rprime|'>\<rangle\>)>.

    \ \ \ \ { \ \ 

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - apply L_catch.

    \ \ \ \ \ \ - \ {

    \ \ \ \ \ \ \ \ \ \ apply T_st_l.

    \ \ \ \ \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_cl_l.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_map_flatten_2.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_k_env_l_2.\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }

    \ \ - intro D2.

    \ \ \ \ inversion D2.

    \ \ \ \ inversion H9.

    \ \ \ \ subst.

    \ \ \ \ assert (H333 := (fetch_mu_sound_l _ _ _ H0 _ H23)).

    \ \ \ \ destruct H333 as [ <math|\<cal-S\>>2 H444 ].

    \ \ \ \ destruct H444.

    \ \ \ \ assert (H555 := (map_flatten_sound _ _ _ _ _ H _ H22)).

    \ \ \ \ destruct H555.

    \ \ \ \ destruct H3.

    \ \ \ \ exists (<math|\<langle\>>t, x, <math|\<cal-L\>>_<math|\<mu\>>,
    <math|\<cal-E\>>_<math|\<mu\>>', <math|\<cal-S\>>2<math|\<rangle\>>).

    \ \ \ \ {\ 

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ apply L_throw.

    \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ apply T_st_l.

    \ \ \ \ \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_cl_l.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }
  </proof>

  <\comment>
    <subsubsection|Completeness>
  </comment>

  <with|font-series|bold|Lemma> <with|font-shape|italic|map_flatten_complete>
  : <math|\<forall\>> {<math|n> <math|\<cal-E\>> <math|\<cal-L\><rprime|'>>
  <math|\<cal-L\><rsub|\<mu\>>> <math|\<alpha\>>},\ 

  <space|2em><math|\<cal-L\><rsub|\<mu\>><protectl|(>\<alpha\><protectr|)><protectlr|=>
  \<cal-L\><rprime|'>> <math|\<rightarrow\>> <math|\<forall\>>
  <math|\<cal-I\><rsub|\<mu\>>>, <with|font-shape|italic|map_flatten>
  <math|n> <math|\<cal-E\>> <math|\<cal-I\><rsub|\<mu\>>>
  <math|\<cal-L\><rsub|\<mu\>>> <math|\<rightarrow\>> <math|\<exists\>>
  <math|\<cal-I\><rprime|'>>, <with|font-shape|italic|flatten> <math|n>
  <math|\<cal-E\>> <math|\<cal-I\><rprime|'>> <math|\<cal-L\><rprime|'>>
  <math|\<wedge\>> <math|\<cal-I\><rsub|\<mu\>><protectl|(>\<alpha\><protectr|)><protectlr|=>
  \<cal-I\><rprime|'>>.

  <\proof>
    \ \ intros n <math|\<cal-E\>> <math|\<cal-L\>>'
    <math|\<cal-L\>>_<math|\<mu\>> <math|\<alpha\>> <math|\<cal-D\>>_2.

    \ \ induction <math|\<cal-D\>>_2.

    \ \ - intros <math|\<cal-I\>>_<math|\<mu\>> <math|\<cal-D\>>_1.

    \ \ \ \ inverts <math|\<cal-D\>>_1 as D11 D12.

    \ \ \ \ exists <math|\<cal-I\>>0.

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - exact D11.

    \ \ \ \ \ \ - apply Fetch_1.

    \ \ \ \ }

    \ \ - intros <math|\<cal-I\>>_<math|\<mu\>> <math|\<cal-D\>>_1.

    \ \ \ \ inverts <math|\<cal-D\>>_1 as D11 D12.

    \ \ \ \ assert (D34 := IH<math|\<cal-D\>>_2 _ D12).

    \ \ \ \ destruct D34 as [<math|\<cal-I\>>' D34].

    \ \ \ \ exists <math|\<cal-I\>>'.

    \ \ \ \ destruct D34 as [ D3 D4 ].

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - exact D3.

    \ \ \ \ \ \ - apply Fetch_2.

    \ \ \ \ \ \ \ \ exact D4.

    \ \ \ \ }
  </proof>

  <with|font-series|bold|Lemma> <with|font-shape|italic|fetch_complete_l> :
  \ <math|\<forall\>> {<math|n> <math|l> <math|c<rprime|'>> <math|\<cal-I\>>
  <math|\<cal-E\>> <math|\<cal-L\>>},\ 

  <space|2em><math|<math-it|flatten><space|1spc>n<space|1spc>\<cal-E\><space|1spc>\<cal-I\><space|1spc>\<cal-L\>\<rightarrow\>\<cal-L\><protectl|(>l<protectr|)>
  <protectlr|=> c<rprime|'>\<rightarrow\>\<exists\><space|1spc>k<space|1spc>c,<space|1spc>\<cal-E\><protectlr|(><math|n>
  <protectl|<math|->> k<protectlr|)> <protectlr|=>
  c<space|1spc>\<wedge\><space|1spc>c<rsup|\<diamond\>>=<rsub|c>
  c<rprime|'>\<wedge\>\<cal-I\><protectl|(>l<protectr|)> <protectlr|=> k>.

  <\proof>
    \ \ intros n l c' <math|\<cal-I\>> <math|\<cal-E\>> <math|\<cal-L\>> D1
    D2.

    \ \ revert D1.

    \ \ revert <math|\<cal-I\>>.

    \ \ induction D2.

    \ \ - intros <math|\<cal-I\>>0 D1.

    \ \ \ \ inverts D1 as D31 D32 D33.

    \ \ \ \ exists k.

    \ \ \ \ exists c.

    \ \ \ \ {\ 

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - exact D31.

    \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ split.

    \ \ \ \ \ \ \ \ \ \ \ \ - exact D32.

    \ \ \ \ \ \ \ \ \ \ \ \ - apply Fetch_1.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }

    \ \ - intros <math|\<cal-I\>>0 D1.

    \ \ \ \ inverts D1 as D31 D32 D33.

    \ \ \ \ assert (D2411 := IHD2 _ D33).

    \ \ \ \ destruct D2411 as [k0 D2411].

    \ \ \ \ destruct D2411 as [c0 D2411].

    \ \ \ \ exists k0 c0.

    \ \ \ \ destruct D2411 as [D20 D411].

    \ \ \ \ destruct D411 as [D4 D11].

    \ \ \ \ {

    \ \ \ \ \ \ split. \ \ \ 

    \ \ \ \ \ \ - exact D20.

    \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ split.

    \ \ \ \ \ \ \ \ \ \ - exact D4.

    \ \ \ \ \ \ \ \ \ \ - apply Fetch_2.

    \ \ \ \ \ \ \ \ \ \ \ \ exact D11.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }
  </proof>

  <with|font-series|bold|Lemma> <with|font-shape|italic|fetch_mu_complete_l>
  : \ <math|\<forall\>> {<math|n> <math|\<cal-S\><rprime|'>>
  <math|\<cal-E\><rsub|\<mu\>><rprime|'>>},\ 

  <space|2em><math|\<cal-E\><rsub|\<mu\>><rprime|'><protectl|(>n<protectr|)>
  <protectlr|=> \<cal-S\><rprime|'>\<rightarrow\>\<forall\>\<cal-E\><rsub|\<mu\>>,\<cal-E\><rsub|\<mu\>><rsup|\<diamond\>>=<rsub|k>
  \<cal-E\><rsub|\<mu\>><rprime|'>\<rightarrow\>\<exists\>\<cal-S\>,\<cal-S\><rsup|\<diamond\>>=<rsub|s>
  \<cal-S\><rprime|'>\<wedge\>\<cal-E\><rsub|\<mu\>><protectl|(>n<protectr|)>
  <protectlr|=> \<cal-S\>>.

  <\proof>
    \ \ intros n S <math|\<cal-E\><rsub|\<mu\>><rprime|'>> D2.

    \ \ induction D2.

    \ \ - intros <math|\<cal-E\><rsub|\<mu\>>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ exists <math|\<cal-S\>>.

    \ \ \ \ {\ 

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - assumption.

    \ \ \ \ \ \ - apply Fetch_1.

    \ \ \ \ }

    \ \ - intros <math|\<cal-E\><rsub|\<mu\>>> D1.

    \ \ \ \ inversion D1.

    \ \ \ \ apply IHD2 in H3.

    \ \ \ \ destruct H3.

    \ \ \ \ destruct H3.

    \ \ \ \ exists x.

    \ \ \ \ {

    \ \ \ \ \ \ split. \ \ \ 

    \ \ \ \ \ \ - exact H3.

    \ \ \ \ \ \ - apply Fetch_2.

    \ \ \ \ \ \ \ \ exact H4.

    \ \ \ \ }
  </proof>

  <with|font-series|bold|Theorem> <with|font-shape|italic|completeness_l> :
  \ <math|\<forall\>> <math|\<sigma\><rsub|1><rprime|'>>
  <math|<with|math-font|cal|\<sigma\><rsub|2>><rprime|'>>
  <math|\<sigma\><rsub|1>>, <math|\<sigma\><rsub|1><rprime|'>
  \<rightsquigarrow\><rsup|l> \<sigma\><rsub|2><rprime|'><space|1spc>\<rightarrow\><space|1spc>\<sigma\><rsub|1><rsup|\<diamond\>>=<rsub|\<sigma\>>
  \<sigma\><rsub|1><rprime|'><space|1spc>\<rightarrow\><space|1spc>\<exists\>\<sigma\><rsub|2>,<space|1spc>\<sigma\><rsub|1>
  \<rightsquigarrow\><rsup|i> \<sigma\><rsub|2><space|1spc>\<wedge\><space|1spc><with|math-font|cal|\<sigma\><rsub|2>><rsup|\<diamond\>>=<rsub|\<sigma\>>
  <with|math-font|cal|\<sigma\><rsub|2><rprime|'>>>.

  <\proof>
    \ \ intros <math|\<sigma\><rsub|1><rprime|'>>
    <math|<with|math-font|cal|\<sigma\><rsub|2>><rprime|'>>
    <math|\<sigma\><rsub|1>> D1.

    \ \ induction D1.

    \ \ - intro D2.

    \ \ \ \ inverts keep D2 as D21 D22.

    \ \ \ \ inverts keep D21 as D211 D212 D213.

    \ \ \ \ assert (H222 := fetch_complete_l D211 H).

    \ \ \ \ destruct H222 as [ k0 H222 ].

    \ \ \ \ destruct H222 as [ c H222 ].

    \ \ \ \ destruct H222 as [ D112_12 D31_111 ].

    \ \ \ \ inverts D112_12 as D112 D12.

    \ \ \ \ destruct D31_111 as [ D31 D111 ].

    \ \ \ \ destruct c as [ t0 n0 <math|\<cal-I\>>0
    <math|\<cal-I\>>_<math|\<mu\>>0 <math|\<cal-E\>>0
    <math|\<cal-E\>>_<math|\<mu\>>1 ].

    \ \ \ \ exists (<math|\<langle\>t0, n0, \<cal-I\>0, \<cal-I\>_\<mu\>0,
    \<cal-E\>0, \<cal-E\>_\<mu\>1, \<cal-S\>0 \<rangle\>>).

    \ \ \ \ {\ 

    \ \ \ \ \ \ split.

    \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ eapply I_var.

    \ \ \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ \ \ eapply Compute_1.

    \ \ \ \ \ \ \ \ \ \ \ \ - exact D111.

    \ \ \ \ \ \ \ \ \ \ \ \ - exact D112.

    \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ - exact D12.

    \ \ \ \ \ \ }

    \ \ \ \ - {

    \ \ \ \ \ \ \ \ apply T_st_l.

    \ \ \ \ \ \ \ \ - exact D31.

    \ \ \ \ \ \ \ \ - exact D22.

    \ \ \ \ \ \ }

    \ \ \ \ }

    \ \ - intro D2.

    \ \ \ \ inverts D2 as D21 D22.

    \ \ \ \ inverts D21 as D211 D212 D213.

    \ \ \ \ exists (<math|\<langle\>t, n, \<cal-I\>, \<cal-I\>_\<mu\>,
    \<cal-E\>, \<cal-E\>_\<mu\>0, [u, n, \<cal-I\>, \<cal-I\>_\<mu\>,
    \<cal-E\>, \<cal-E\>_\<mu\>0]::\<cal-S\>0\<rangle\>>).

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - apply I_app.

    \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ apply T_st_l.

    \ \ \ \ \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_cl_l.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D211.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D212.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D213.

    \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_stack_l_2.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ {

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_cl_l.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D211.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D212.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D213.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D22.

    \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }

    \ \ - intro D2.

    \ \ \ \ inverts D2 as D21 D22.

    \ \ \ \ inverts D21 as D211 D212 D213.

    \ \ \ \ inverts D22 as D221 D222.

    \ \ \ \ exists (<math|\<langle\>t, (S n), (S n)::\<cal-I\>,
    \<cal-I\>_\<mu\>, (c0::\<cal-E\>), \<cal-E\>_\<mu\>0,
    \<cal-S\>0\<rangle\>>).

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - apply I_abs.

    \ \ \ \ \ \ - assert (D1 := minus_id (S n)).\ 

    \ \ \ \ \ \ \ \ assert (D2 := weaken_flatten c0 D211).

    \ \ \ \ \ \ \ \ assert (D3 := weaken_map_flatten c0 D212).

    \ \ \ \ \ \ \ \ {\ 

    \ \ \ \ \ \ \ \ \ \ apply T_st_l.

    \ \ \ \ \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_cl_l.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ eapply T_flatten_2.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - {
    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ eapply Compute_l_1.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D1.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - apply Fetch_1.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D221.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D2.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D3.\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D213.

    \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ - exact D222.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }

    \ \ - intro D2.

    \ \ \ \ inverts D2 as D21 D22.

    \ \ \ \ inverts D21 as D211 D212 D213.

    \ \ \ \ exists (<math|\<langle\>t, n, \<cal-I\>,
    (\<cal-I\>::\<cal-I\>_\<mu\>), \<cal-E\>,
    (\<cal-S\>0::\<cal-E\>_\<mu\>0), \<cal-S\>0\<rangle\>>).

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - apply I_catch.

    \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ apply T_st_l.

    \ \ \ \ \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_cl_l.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D211.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ eapply T_map_flatten_2.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - \ exact D211.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - \ exact D212.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_k_env_l_2.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D22.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D213.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ - exact D22.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }

    \ \ - intro D2.

    \ \ \ \ inverts D2 as D21 D22.

    \ \ \ \ inverts D21 as D211 D212 D213.

    \ \ \ \ assert (D32_12 := fetch_mu_complete_l H0 _ D213).

    \ \ \ \ assert (D5_11 := map_flatten_complete H _ D212).

    \ \ \ \ destruct D32_12 as [ <math|\<cal-S\>>1 D32_12 ].

    \ \ \ \ destruct D32_12 as [ D32 D12 ].

    \ \ \ \ destruct D5_11 as [ <math|\<cal-I\>>1 D5_11 ].

    \ \ \ \ destruct D5_11 as [ D5 D11 ].

    \ \ \ \ exists (<math|\<langle\>t, n, \<cal-I\>1, \<cal-I\>_\<mu\>,
    \<cal-E\>, \<cal-E\>_\<mu\>0, \<cal-S\>1\<rangle\>>).

    \ \ \ \ {

    \ \ \ \ \ \ split.

    \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ apply I_throw.

    \ \ \ \ \ \ \ \ \ \ - exact D11.

    \ \ \ \ \ \ \ \ \ \ - exact D12.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ - {

    \ \ \ \ \ \ \ \ \ \ apply T_st_l.

    \ \ \ \ \ \ \ \ \ \ - {\ 

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ apply T_cl_l.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D5.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D212.

    \ \ \ \ \ \ \ \ \ \ \ \ \ \ - exact D213.

    \ \ \ \ \ \ \ \ \ \ \ \ }

    \ \ \ \ \ \ \ \ \ \ - exact D32.

    \ \ \ \ \ \ \ \ }

    \ \ \ \ }
  </proof>

  <\comment>
    <section|Conclusion and future work>
  </comment>
</body>

<\initial>
  <\collection>
    <associate|font-base-size|10>
    <associate|page-screen-margin|false>
    <associate|par-first|0tab>
  </collection>
</initial>

<\references>
  <\collection>
    <associate|auto-1|<tuple|1|5>>
    <associate|auto-10|<tuple|3.4.1|16>>
    <associate|auto-11|<tuple|3.4.2|17>>
    <associate|auto-12|<tuple|4|20>>
    <associate|auto-13|<tuple|4.1|24>>
    <associate|auto-14|<tuple|4.1.1|?>>
    <associate|auto-15|<tuple|4.1.2|?>>
    <associate|auto-16|<tuple|4.2|?>>
    <associate|auto-17|<tuple|4.2.1|?>>
    <associate|auto-18|<tuple|4.2.2|?>>
    <associate|auto-19|<tuple|4.3|?>>
    <associate|auto-2|<tuple|2|5>>
    <associate|auto-20|<tuple|4.3.1|?>>
    <associate|auto-21|<tuple|4.3.2|?>>
    <associate|auto-22|<tuple|5|?>>
    <associate|auto-3|<tuple|3|7>>
    <associate|auto-4|<tuple|3.1|7>>
    <associate|auto-5|<tuple|3.2|7>>
    <associate|auto-6|<tuple|3.3|8>>
    <associate|auto-7|<tuple|3.3.1|9>>
    <associate|auto-8|<tuple|3.3.2|10>>
    <associate|auto-9|<tuple|3.4|13>>
  </collection>
</references>

<\auxiliary>
  <\collection>
    <\associate|toc>
      <vspace*|1fn><with|font-series|<quote|bold>|math-font-series|<quote|bold>|1<space|2spc>Introduction>
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-1><vspace|0.5fn>

      <vspace*|1fn><with|font-series|<quote|bold>|math-font-series|<quote|bold>|2<space|2spc>Dependency
      relations> <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-2><vspace|0.5fn>

      <vspace*|1fn><with|font-series|<quote|bold>|math-font-series|<quote|bold>|3<space|2spc>Abstract
      machines> <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-3><vspace|0.5fn>

      <with|par-left|<quote|1tab>|3.1<space|2spc>Safe
      <with|mode|<quote|math>|\<lambda\>><rsub|ct>-terms
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-4>>

      <with|par-left|<quote|1tab>|3.2<space|2spc>From local indices to global
      indices <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-5>>

      <with|par-left|<quote|1tab>|3.3<space|2spc>Abstract machine for
      <with|mode|<quote|math>|\<lambda\><rsub|<rigid|<with|mode|<quote|text>|<with|font-family|<quote|rm>|font-series|<quote|bold>|font-shape|<quote|right>|ct>>>>>-terms
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-6>>

      <with|par-left|<quote|2tab>|3.3.1<space|2spc>Closure, environment,
      stack and state <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-7>>

      <with|par-left|<quote|2tab>|3.3.2<space|2spc>Evaluation rules
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-8>>

      <with|par-left|<quote|1tab>|3.4<space|2spc>Abstract machine for safe
      <with|mode|<quote|math>|\<lambda\><rsub|<rigid|<with|mode|<quote|text>|<with|font-family|<quote|rm>|font-series|<quote|bold>|font-shape|<quote|right>|ct>>>>>-terms
      (with local environments) <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-9>>

      <with|par-left|<quote|2tab>|3.4.1<space|2spc>Closure, environment,
      stack and state <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-10>>

      <with|par-left|<quote|2tab>|3.4.2<space|2spc>Evaluation rules
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-11>>

      <vspace*|1fn><with|font-series|<quote|bold>|math-font-series|<quote|bold>|4<space|2spc>Bisimulations>
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-12><vspace|0.5fn>

      <with|par-left|<quote|1tab>|4.1<space|2spc>Abstract machine for safe
      <with|mode|<quote|math>|\<lambda\><rsub|<rigid|<with|mode|<quote|text>|<with|font-family|<quote|rm>|font-series|<quote|bold>|font-shape|<quote|right>|ct>>>>>-terms
      (with indirection tables) <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-13>>

      <with|par-left|<quote|2tab>|4.1.1<space|2spc>Closure, environment,
      stack and state <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-14>>

      <with|par-left|<quote|2tab>|4.1.2<space|2spc>Evaluation rules
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-15>>

      <with|par-left|<quote|1tab>|4.2<space|2spc>Lock-step simulation
      <with|mode|<quote|math>|<around*|(|-|)><rsup|\<star\>>>
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-16>>

      <with|par-left|<quote|2tab>|4.2.1<space|2spc>Soundness
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-17>>

      <with|par-left|<quote|2tab>|4.2.2<space|2spc>Completeness
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-18>>

      <with|par-left|<quote|1tab>|4.3<space|2spc>Lock-step simulation
      <with|mode|<quote|math>|<around*|(|-|)><rsup|\<diamond\>>>
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-19>>

      <with|par-left|<quote|2tab>|4.3.1<space|2spc>Soundness
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-20>>

      <with|par-left|<quote|2tab>|4.3.2<space|2spc>Completeness
      <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-21>>

      <vspace*|1fn><with|font-series|<quote|bold>|math-font-series|<quote|bold>|5<space|2spc>Conclusion
      and future work> <datoms|<macro|x|<repeat|<arg|x>|<with|font-series|medium|<with|font-size|1|<space|0.2fn>.<space|0.2fn>>>>>|<htab|5mm>>
      <no-break><pageref|auto-22><vspace|0.5fn>
    </associate>
  </collection>
</auxiliary>